// Copyright Notice
// ================
// BOCHS is Copyright 1994-1999 by Kevin P. Lawton.
//
// BOCHS is commercial software.
//
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
//
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02452-3028
//
//     EMail:    bochs@world.std.com





#include "bochs.h"

#define Repeatable      0x0040
#define RepeatableZF    0x0080


#if BX_SHOW_IPS
unsigned long ips_count=0;
#endif


//unsigned counter[2] = { 0, 0 };



#if BX_SIM_ID == 0   // only need to define once
// This array defines a look-up table for the even parity-ness
// of an 8bit quantity, for optimal assignment of the parity bit
// in the EFLAGS register
const Boolean bx_parity_lookup[256] = {
  1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
  0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
  0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
  1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
  0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
  1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
  1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
  0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
  0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
  1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
  1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
  0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
  1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
  0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
  0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
  1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1
  };
#endif


BX_CPU_C    BX_CPU;
BX_MEM_C    BX_MEM;





// notes:
//
// check limit of CS?


  void
BX_CPU_C::cpu_loop(void)
{
  unsigned ret;
  BxInstruction_t i;
  unsigned maxisize;
  Bit8u *fetch_ptr;
  Boolean is_32;


  (void) setjmp( BX_CPU_THIS_PTR jmp_buf_env );


loopy:
#if BX_DEBUGGER
    {
    Bit32u debug_eip = BX_CPU_THIS_PTR prev_eip;
    if ( dbg_is_begin_instr_bpoint(
           BX_CPU_THIS_PTR cs.selector.value,
           debug_eip,
           BX_CPU_THIS_PTR cs.cache.u.segment.base + debug_eip,
           BX_CPU_THIS_PTR cs.cache.u.segment.d_b) ) {
      return;
      }
    }
#endif  // #if BX_DEBUGGER

// ???
BX_CPU_THIS_PTR EXT = 0;
BX_CPU_THIS_PTR errorno = 0;

  if (BX_CPU_THIS_PTR bytesleft == 0) {
    prefetch();
    }
  fetch_ptr = BX_CPU_THIS_PTR fetch_ptr;

  maxisize = 16;
  if (BX_CPU_THIS_PTR bytesleft < 16)
    maxisize = BX_CPU_THIS_PTR bytesleft;
  is_32 = BX_CPU_THIS_PTR cs.cache.u.segment.d_b;
  ret = FetchDecode(fetch_ptr, &i, maxisize, is_32);

  if (ret) {
    if (i.ResolveModrmAddr) {
      i.ResolveModrmAddr(&i);
      }
    BX_CPU_THIS_PTR fetch_ptr += i.ilen;
    BX_CPU_THIS_PTR bytesleft -= i.ilen;
fetch_decode_OK:

    if (i.rep_used && (i.attr & Repeatable)) {
repeat_loop:
      if (i.attr & RepeatableZF) {
        if (i.as_32) {
          if (ECX != 0) {
            i.execute(&i);
#if BX_SHOW_IPS
            ips_count++;
#endif
            ECX -= 1;
            }
          if ((i.rep_used==0xf3) && (get_ZF()==0)) goto repeat_done;
          if ((i.rep_used==0xf2) && (get_ZF()!=0)) goto repeat_done;
          if (ECX == 0) goto repeat_done;
          goto repeat_not_done;
          }
        else {
          if (CX != 0) {
            i.execute(&i);
#if BX_SHOW_IPS
            ips_count++;
#endif
            CX -= 1;
            }
          if ((i.rep_used==0xf3) && (get_ZF()==0)) goto repeat_done;
          if ((i.rep_used==0xf2) && (get_ZF()!=0)) goto repeat_done;
          if (CX == 0) goto repeat_done;
          goto repeat_not_done;
          }
        }
      else { // normal repeat, no concern for ZF
        if (i.as_32) {
          if (ECX != 0) {
            i.execute(&i);
#if BX_SHOW_IPS
            ips_count++;
#endif
            ECX -= 1;
            }
          if (ECX == 0) goto repeat_done;
          goto repeat_not_done;
          }
        else { // 16bit addrsize
          if (CX != 0) {
            i.execute(&i);
#if BX_SHOW_IPS
            ips_count++;
#endif
            CX -= 1;
            }
          if (CX == 0) goto repeat_done;
          goto repeat_not_done;
          }
        }
      // shouldn't get here from above
repeat_not_done:
      BX_TICK1();
#if BX_DEBUGGER == 0
      if (BX_CPU_THIS_PTR async_event) {
#endif
        invalidate_prefetch_q();
        goto debugger_check;
#if BX_DEBUGGER == 0
        }
#endif
      goto repeat_loop;

repeat_done:
      BX_CPU_THIS_PTR eip += i.ilen;
      }
    else {
      BX_CPU_THIS_PTR eip += i.ilen;
      i.execute(&i);
#if BX_SHOW_IPS
      ips_count++;
#endif
      }

    BX_CPU_THIS_PTR prev_eip = EIP; // commit new EIP
    BX_CPU_THIS_PTR prev_esp = ESP; // commit new ESP
    BX_TICK1();
debugger_check:
#if BX_DEBUGGER
    {
    Bit32u debug_eip = BX_CPU_THIS_PTR prev_eip;
    if ( dbg_is_end_instr_bpoint(
           BX_CPU_THIS_PTR cs.selector.value,
           debug_eip,
           BX_CPU_THIS_PTR cs.cache.u.segment.base + debug_eip,
           BX_CPU_THIS_PTR cs.cache.u.segment.d_b) ) {
      return;
      }
    }
#endif  // #if BX_DEBUGGER
    if (BX_CPU_THIS_PTR async_event)
      goto handle_async_event;
    goto loopy;
    }
  else {
    unsigned remain, j;
static Bit8u FetchBuffer[16];
    Bit8u *temp_ptr;

    // read all leftover bytes in current page
    for (j=0; j<BX_CPU_THIS_PTR bytesleft; j++) {
      FetchBuffer[j] = *fetch_ptr++;
      }

    // get remaining bytes for prefetch in next page
    // prefetch() needs eip current
    BX_CPU_THIS_PTR eip += BX_CPU_THIS_PTR bytesleft;
    remain = BX_CPU_THIS_PTR bytesleft;
    prefetch();

    if (BX_CPU_THIS_PTR bytesleft < 16) {
      // make sure (bytesleft - remain) below doesn't go negative
      bx_panic("fetch_decode: bytesleft==0 after prefetch\n");
      }
    temp_ptr = fetch_ptr = BX_CPU_THIS_PTR fetch_ptr;

    // read leftover bytes in next page
    for (; j<16; j++) {
      FetchBuffer[j] = *temp_ptr++;
      }
    ret = FetchDecode(FetchBuffer, &i, 16, is_32);
    if (ret==0)
      bx_panic("fetchdecode: cross boundary: ret==0\n");
    if (i.ResolveModrmAddr) {
      i.ResolveModrmAddr(&i);
      }
    remain = i.ilen - remain;

    // note: eip has already been advanced to beginning of page
    BX_CPU_THIS_PTR fetch_ptr = fetch_ptr + remain;
    BX_CPU_THIS_PTR bytesleft -= remain;
    //BX_CPU_THIS_PTR eip += remain;
    BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip;
    goto fetch_decode_OK;
    }


handle_async_event:
  if (BX_CPU_THIS_PTR single_step_event) {
    bx_panic("async_event single step event\n");
    }
  if (BX_CPU_THIS_PTR eflags.tf) {
    bx_panic("async_event TF\n");
    }
  if (BX_CPU_THIS_PTR inhibit_interrupts) {
    BX_CPU_THIS_PTR inhibit_interrupts = 0;
    }
  else if (BX_INTR && BX_CPU_THIS_PTR eflags.if_ && BX_DBG_ASYNC_INTR) {
    Bit8u vector;

    // NOTE: similar code in ::take_irq()
    vector = BX_IAC(); // may set INTR with next interrupt
    //if (bx_dbg.interrupts) bx_printf("decode: interrupt %u\n",
    //                                   (unsigned) vector);
    BX_CPU_THIS_PTR errorno = 0;
    BX_CPU_THIS_PTR EXT   = 1; /* external event */
    interrupt(vector, 0, 0, 0);
    BX_INSTR_HWINTERRUPT(vector, BX_CPU_THIS_PTR cs.selector.value, BX_CPU_THIS_PTR eip);
    }
  else if (BX_HRQ && BX_DBG_ASYNC_DMA) {
    // NOTE: similar code in ::take_dma()
    // assert Hold Acknowledge (HLDA) and go into a bus hold state
    BX_RAISE_HLDA();
    }
  if ( !(BX_INTR ||
         BX_CPU_THIS_PTR single_step_event ||
         BX_HRQ ||
         BX_CPU_THIS_PTR eflags.tf) )
    BX_CPU_THIS_PTR async_event = 0;
  goto loopy;
}


#if BX_DEBUGGER
  Boolean
BX_CPU_C::dbg_is_begin_instr_bpoint(Bit32u cs, Bit32u eip, Bit32u laddr,
                                    Bit32u is_32)
{
  bx_guard_found[BX_SIM_ID].cs  = cs;
  bx_guard_found[BX_SIM_ID].eip = eip;
  bx_guard_found[BX_SIM_ID].laddr = laddr;
  bx_guard_found[BX_SIM_ID].is_32bit_code = is_32;

  // see if debugger is looking for iaddr breakpoint of any type
  if (bx_guard.guard_for & BX_DBG_GUARD_IADDR_ALL) {
#if BX_DBG_SUPPORT_VIR_BPOINT
    if (bx_guard.guard_for & BX_DBG_GUARD_IADDR_VIR) {
      if (bx_guard_found[BX_SIM_ID].icount!=0) {
        for (unsigned i=0; i<bx_guard.iaddr.num_virtual; i++) {
          if ( (bx_guard.iaddr.vir[i].cs  == cs) &&
               (bx_guard.iaddr.vir[i].eip == eip) ) {
            bx_guard_found[BX_SIM_ID].guard_found = BX_DBG_GUARD_IADDR_VIR;
            bx_guard_found[BX_SIM_ID].iaddr_index = i;
            return(1); // on a breakpoint
            }
          }
        }
      }
#endif
#if BX_DBG_SUPPORT_LIN_BPOINT
    if (bx_guard.guard_for & BX_DBG_GUARD_IADDR_LIN) {
      if (bx_guard_found[BX_SIM_ID].icount!=0) {
        for (unsigned i=0; i<bx_guard.iaddr.num_linear; i++) {
          if ( bx_guard.iaddr.lin[i].addr == bx_guard_found[BX_SIM_ID].laddr ) {
            bx_guard_found[BX_SIM_ID].guard_found = BX_DBG_GUARD_IADDR_LIN;
            bx_guard_found[BX_SIM_ID].iaddr_index = i;
            return(1); // on a breakpoint
            }
          }
        }
      }
#endif
#if BX_DBG_SUPPORT_PHY_BPOINT
    if (bx_guard.guard_for & BX_DBG_GUARD_IADDR_PHY) {
      Bit32u phy;
      Boolean valid;
      dbg_xlate_linear2phy(bx_guard_found[BX_SIM_ID].laddr,
                              &phy, &valid);
      if ( (bx_guard_found[BX_SIM_ID].icount!=0) && valid ) {
        for (unsigned i=0; i<bx_guard.iaddr.num_physical; i++) {
          if ( bx_guard.iaddr.phy[i].addr == phy ) {
            bx_guard_found[BX_SIM_ID].guard_found = BX_DBG_GUARD_IADDR_PHY;
            bx_guard_found[BX_SIM_ID].iaddr_index = i;
            return(1); // on a breakpoint
            }
          }
        }
      }
#endif
    }
  return(0); // not on a breakpoint
}


  Boolean
BX_CPU_C::dbg_is_end_instr_bpoint(Bit32u cs, Bit32u eip, Bit32u laddr,
                                  Bit32u is_32)
{
  bx_guard_found[BX_SIM_ID].icount++;

  // see if debugger requesting icount guard
  if (bx_guard.guard_for & BX_DBG_GUARD_ICOUNT) {
    if (bx_guard_found[BX_SIM_ID].icount >= bx_guard.icount) {
      bx_guard_found[BX_SIM_ID].cs  = cs;
      bx_guard_found[BX_SIM_ID].eip = eip;
      bx_guard_found[BX_SIM_ID].laddr = laddr;
      bx_guard_found[BX_SIM_ID].is_32bit_code = is_32;
      bx_guard_found[BX_SIM_ID].guard_found = BX_DBG_GUARD_ICOUNT;
      return(1);
      }
    }

  // convenient point to see if user typed Ctrl-C
  if (bx_guard.interrupt_requested &&
      (bx_guard.guard_for & BX_DBG_GUARD_CTRL_C)) {
    bx_guard_found[BX_SIM_ID].guard_found = BX_DBG_GUARD_CTRL_C;
    return(1);
    }

#if (BX_NUM_SIMULATORS >= 2)
  // if async event pending, acknowlege them
  if (bx_guard.async_changes_pending.which) {
    if (bx_guard.async_changes_pending.which & BX_DBG_ASYNC_PENDING_A20)
      bx_dbg_async_pin_ack(BX_DBG_ASYNC_PENDING_A20,
                           bx_guard.async_changes_pending.a20);
    if (bx_guard.async_changes_pending.which) {
      bx_panic("decode: async pending unrecognized.\n");
      }
    }
#endif
  return(0); // no breakpoint
}


  void
BX_CPU_C::dbg_take_irq(void)
{
  unsigned vector;

  // NOTE: similar code in ::cpu_loop()

  if ( BX_INTR && BX_CPU_THIS_PTR eflags.if_ ) {
    if ( setjmp(BX_CPU_THIS_PTR jmp_buf_env) == 0 ) {
      // normal return from setjmp setup
      vector = BX_IAC(); // may set INTR with next interrupt
      BX_CPU_THIS_PTR errorno = 0;
      BX_CPU_THIS_PTR EXT   = 1; // external event
      BX_CPU_THIS_PTR async_event = 1; // set in case INTR is triggered
      interrupt(vector, 0, 0, 0);
      }
    }
}

  void
BX_CPU_C::dbg_force_interrupt(unsigned vector)
{
  // Used to force slave simulator to take an interrupt, without
  // regard to IF

  if ( setjmp(BX_CPU_THIS_PTR jmp_buf_env) == 0 ) {
    // normal return from setjmp setup
    BX_CPU_THIS_PTR errorno = 0;
    BX_CPU_THIS_PTR EXT   = 1; // external event
    BX_CPU_THIS_PTR async_event = 1; // probably don't need this
    interrupt(vector, 0, 0, 0);
    }
}

  void
BX_CPU_C::dbg_take_dma(void)
{
  // NOTE: similar code in ::cpu_loop()
  if ( BX_HRQ ) {
    BX_CPU_THIS_PTR async_event = 1; // set in case INTR is triggered
    BX_RAISE_HLDA();
    }
}
#endif  // #if BX_DEBUGGER
