// Copyright Notice
// ================
// BOCHS is Copyright 1994-1999 by Kevin P. Lawton.
//
// BOCHS is commercial software.
//
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
//
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02452-3028
//
//     EMail:    bochs@world.std.com




#ifndef BX_CPU_H
#  define BX_CPU_H 1

#include <setjmp.h>

#include "cpu/lazy_flags.h"



/* segment register encoding */
#define BX_SEG_REG_ES    0
#define BX_SEG_REG_CS    1
#define BX_SEG_REG_SS    2
#define BX_SEG_REG_DS    3
#define BX_SEG_REG_FS    4
#define BX_SEG_REG_GS    5



#ifdef BX_LITTLE_ENDIAN
#define BX_REG8L_OFFSET 0
#define BX_REG8H_OFFSET 1
#define BX_REG16_OFFSET 0
#else // BX_BIG_ENDIAN
#define BX_REG8L_OFFSET 3
#define BX_REG8H_OFFSET 2
#define BX_REG16_OFFSET 2
#endif // ifdef BX_LITTLE_ENDIAN

#define BX_8BIT_REG_AL  0
#define BX_8BIT_REG_CL  1
#define BX_8BIT_REG_DL  2
#define BX_8BIT_REG_BL  3
#define BX_8BIT_REG_AH  4
#define BX_8BIT_REG_CH  5
#define BX_8BIT_REG_DH  6
#define BX_8BIT_REG_BH  7

#define BX_16BIT_REG_AX 0
#define BX_16BIT_REG_CX 1
#define BX_16BIT_REG_DX 2
#define BX_16BIT_REG_BX 3
#define BX_16BIT_REG_SP 4
#define BX_16BIT_REG_BP 5
#define BX_16BIT_REG_SI 6
#define BX_16BIT_REG_DI 7

#define BX_32BIT_REG_EAX 0
#define BX_32BIT_REG_ECX 1
#define BX_32BIT_REG_EDX 2
#define BX_32BIT_REG_EBX 3
#define BX_32BIT_REG_ESP 4
#define BX_32BIT_REG_EBP 5
#define BX_32BIT_REG_ESI 6
#define BX_32BIT_REG_EDI 7


// access to 8 bit general registers
#define AL (BX_CPU_THIS_PTR gen_reg[0].word.byte.rl)
#define CL (BX_CPU_THIS_PTR gen_reg[1].word.byte.rl)
#define DL (BX_CPU_THIS_PTR gen_reg[2].word.byte.rl)
#define BL (BX_CPU_THIS_PTR gen_reg[3].word.byte.rl)
#define AH (BX_CPU_THIS_PTR gen_reg[0].word.byte.rh)
#define CH (BX_CPU_THIS_PTR gen_reg[1].word.byte.rh)
#define DH (BX_CPU_THIS_PTR gen_reg[2].word.byte.rh)
#define BH (BX_CPU_THIS_PTR gen_reg[3].word.byte.rh)


// access to 16 bit general registers
#define AX (BX_CPU_THIS_PTR gen_reg[0].word.rx)
#define CX (BX_CPU_THIS_PTR gen_reg[1].word.rx)
#define DX (BX_CPU_THIS_PTR gen_reg[2].word.rx)
#define BX (BX_CPU_THIS_PTR gen_reg[3].word.rx)
#define SP (BX_CPU_THIS_PTR gen_reg[4].word.rx)
#define BP (BX_CPU_THIS_PTR gen_reg[5].word.rx)
#define SI (BX_CPU_THIS_PTR gen_reg[6].word.rx)
#define DI (BX_CPU_THIS_PTR gen_reg[7].word.rx)

// access to 16 bit instruction pointer
#define IP (* (Bit16u *) (((Bit8u *) &BX_CPU_THIS_PTR eip) + BX_REG16_OFFSET))


// accesss to 32 bit general registers
#define EAX BX_CPU_THIS_PTR gen_reg[0].erx
#define ECX BX_CPU_THIS_PTR gen_reg[1].erx
#define EDX BX_CPU_THIS_PTR gen_reg[2].erx
#define EBX BX_CPU_THIS_PTR gen_reg[3].erx
#define ESP BX_CPU_THIS_PTR gen_reg[4].erx
#define EBP BX_CPU_THIS_PTR gen_reg[5].erx
#define ESI BX_CPU_THIS_PTR gen_reg[6].erx
#define EDI BX_CPU_THIS_PTR gen_reg[7].erx

// access to 32 bit instruction pointer
#define EIP BX_CPU_THIS_PTR eip

#define BX_READ_8BIT_REG(index)  (((index) < 4) ? \
  (BX_CPU_THIS_PTR gen_reg[index].word.byte.rl) : \
  (BX_CPU_THIS_PTR gen_reg[(index)-4].word.byte.rh))
#define BX_READ_16BIT_REG(index) (BX_CPU_THIS_PTR gen_reg[index].word.rx)
#define BX_READ_32BIT_REG(index) (BX_CPU_THIS_PTR gen_reg[index].erx)

#define BX_READ_16BIT_BASE_REG(var, index) {\
  var = *BX_CPU_THIS_PTR _16bit_base_reg[index];\
  }

#define BX_READ_16BIT_INDEX_REG(var, index) {\
  var = *BX_CPU_THIS_PTR _16bit_index_reg[index];\
  }

#define BX_WRITE_8BIT_REG(index, val) {\
  if ((index) < 4) \
    BX_CPU_THIS_PTR gen_reg[index].word.byte.rl = val; \
  else \
    BX_CPU_THIS_PTR gen_reg[(index)-4].word.byte.rh = val; \
  }
#define BX_WRITE_16BIT_REG(index, val) {\
  BX_CPU_THIS_PTR gen_reg[index].word.rx = val; \
  }
#define BX_WRITE_32BIT_REG(index, val) {\
  BX_CPU_THIS_PTR gen_reg[index].erx = val; \
  }




#define TF BX_CPU_THIS_PTR eflags.tf
#define IF BX_CPU_THIS_PTR eflags.if_
#define DF BX_CPU_THIS_PTR eflags.df

#define IOPL BX_CPU_THIS_PTR eflags.iopl
#define CPL  BX_CPU_THIS_PTR cs.selector.rpl

#define BX_DE_EXCEPTION   0     /* Divide Error exception */
#define BX_UD_EXCEPTION   6
#define BX_NM_EXCEPTION   7
#define BX_DF_EXCEPTION   8
#define BX_TS_EXCEPTION  10
#define BX_NP_EXCEPTION  11
#define BX_SS_EXCEPTION  12
#define BX_GP_EXCEPTION  13
#define BX_PF_EXCEPTION  14
#define BX_MF_EXCEPTION  16
#define BX_AC_EXCEPTION  17





typedef struct {
  /* 31|30|29|28|27|26|25|24|23|22|21|20|19|18|17|16
   * ==|==|=====|==|==|==|==|==|==|==|==|==|==|==|==
   *  0| 0| 0| 0| 0| 0| 0| 0| 0| 0|ID|VP|VF|AC|VM|RF
   *
   * 15|14|13|12|11|10| 9| 8| 7| 6| 5| 4| 3| 2| 1| 0
   * ==|==|=====|==|==|==|==|==|==|==|==|==|==|==|==
   *  0|NT| IOPL|OF|DF|IF|TF|SF|ZF| 0|AF| 0|PF| 1|CF
   */

  Boolean cf;
  Boolean bit1;
  Bit8u   pf_byte;  /* PF derived from last result byte when needed */
  Boolean bit3;
  Boolean af;
  Boolean bit5;
  Boolean zf;
  Boolean sf;
  Boolean tf;
  Boolean if_;
  Boolean df;
  Boolean of;
#if BX_CPU_LEVEL >= 2
  Bit8u   iopl;
  Boolean nt;
#endif
  Boolean bit15;
#if BX_CPU_LEVEL >= 3
  Boolean rf;
  Boolean vm;
#endif
#if BX_CPU_LEVEL >= 4
  Boolean ac;
#endif
  } bx_flags_reg_t;


#if BX_CPU_LEVEL >= 2
typedef struct {
  Bit32u  val32; // 32bit value of register

  // bitfields broken out for efficient access
#if BX_CPU_LEVEL >= 3
  Boolean pg; // paging
#endif

// CR0 notes:
//   Each x86 level has its own quirks regarding how it handles
//   reserved bits.  I used DOS DEBUG.EXE in real mode on the
//   following processors, tried to clear bits 5..30, then tried
//   to set bits 5..30, to see how these bits are handled.
//   I found the following:
//
//   Processor    try to clear bits 5..30    try to set bits 5..30
//   386          7FFFFFF0                   7FFFFFF0
//   486          ????????                   ????????
//   Pentium      00000010                   7FFFFFF0
//   Pentium-II   00000010                   60050030
//
// My assumptions:
//   All processors: bit 5 is hardwired to 1 (not true on all clones)
//   386: bits 6..30 of CR0 are also hardwired to 1
//   Pentium: reserved bits retain value set using mov cr0, reg32
//   Pentium-II: reserved bits are hardwired to 0

#if BX_CPU_LEVEL >= 4
  /* uncomment initialization in main.c */
  Boolean cd; // cache disable
  Boolean nw; // no write-through
  Boolean am; // alignment mask
  Boolean wp; // write-protect
  Boolean ne; // numerics exception
#endif

  Boolean ts; // task switched
  Boolean em; // emulate math coprocessor
  Boolean mp; // monitor coprocessor
  Boolean pe; // protected mode enable
  } bx_cr0_t;
#endif


typedef struct { /* bx_selector_t */
  Bit16u value;   /* the 16bit value of the selector */
#if BX_CPU_LEVEL >= 2
    /* the following fields are extracted from the value field in protected
       mode only.  They're used for sake of efficiency */
  Bit16u index;   /* 13bit index extracted from value in protected mode */
  Bit8u  ti;      /* table indicator bit extracted from value */
  Bit8u  rpl;     /* RPL extracted from value */
#endif
  } bx_selector_t;



typedef struct {
  Boolean valid;         /* 0 = invalid, 1 = valid */
  Boolean p;             /* present */
  Bit8u   dpl;           /* descriptor privilege level 0..3 */
  Boolean segment;       /* 0 = system/gate, 1 = data/code segment */
  Bit8u   type;          /* For system & gate descriptors, only
                          *  0 = invalid descriptor (reserved)
                          *  1 = 286 available Task State Segment (TSS)
                          *  2 = LDT descriptor
                          *  3 = 286 busy Task State Segment (TSS)
                          *  4 = 286 call gate
                          *  5 = task gate
                          *  6 = 286 interrupt gate
                          *  7 = 286 trap gate
                          *  8 = (reserved)
                          *  9 = 386 available TSS
                          * 10 = (reserved)
                          * 11 = 386 busy TSS
                          * 12 = 386 call gate
                          * 13 = (reserved)
                          * 14 = 386 interrupt gate
                          * 15 = 386 trap gate */
  union {
  struct {
    Boolean executable;    /* 1=code, 0=data or stack segment */
    Boolean c_ed;          /* for code: 1=conforming,
                              for data/stack: 1=expand down */
    Boolean r_w;           /* for code: readable?, for data/stack: writeable? */
    Boolean a;             /* accessed? */
    Bit32u  base;          /* base address: 286=24bits, 386=32bits */
    Bit32u  limit;         /* limit: 286=16bits, 386=20bits */
    Bit32u  limit_scaled;  /* for efficiency, this contrived field is set to
                            * limit for byte granular, and
                            * (limit << 12) | 0xfff for page granular seg's
                            */
#if BX_CPU_LEVEL >= 3
    Boolean g;             /* granularity: 0=byte, 1=4K (page) */
    Boolean d_b;           /* default size: 0=16bit, 1=32bit */
    Boolean avl;           /* available for use by system */
#endif
    } segment;
  struct {
    Bit8u   word_count;    /* 5bits (0..31) #words to copy from caller's stack
                            * to called procedure's stack.  (call gates only)*/
    Bit16u  dest_selector;
    Bit16u  dest_offset;
    } gate286;
  struct {                 // type 5: Task Gate Descriptor
    Bit16u  tss_selector;  // TSS segment selector
    } taskgate;
#if BX_CPU_LEVEL >= 3
  struct {
    Bit8u   dword_count;   /* 5bits (0..31) #dwords to copy from caller's stack
                            * to called procedure's stack.  (call gates only)*/
    Bit16u  dest_selector;
    Bit32u  dest_offset;
    } gate386;
#endif
  struct {
    Bit32u  base;          /* 24 bit 286 TSS base  */
    Bit16u  limit;         /* 16 bit 286 TSS limit */
    } tss286;
#if BX_CPU_LEVEL >= 3
  struct {
    Bit32u  base;          /* 32 bit 386 TSS base  */
    Bit32u  limit;         /* 20 bit 386 TSS limit */
    Boolean g;             /* granularity: 0=byte, 1=4K (page) */
    Boolean avl;           /* available for use by system */
    } tss386;
#endif
  struct {
    Bit32u  base;  /* 286=24 386+ =32 bit LDT base */
    Bit16u  limit; /* 286+ =16 bit LDT limit */
    } ldt;
    } u;

  } bx_descriptor_t;

typedef struct {
  bx_selector_t          selector;
  bx_descriptor_t  cache;
  } bx_segment_reg_t;


typedef struct BxInstruction_tag {
  // prefix stuff here...
  unsigned attr; // attribute from fetchdecode
  unsigned b1; // opcode1 byte
  unsigned rep_used;
  unsigned modrm; // mod-nnn-r/m byte
    unsigned mod;
    unsigned nnn;
    unsigned rm;
  Bit16u displ16u; // for 16-bit modrm forms
  Bit32u displ32u; // for 32-bit modrm forms
  bx_segment_reg_t *seg_reg;
  unsigned sib; // scale-index-base (2nd modrm byte)
    unsigned scale;
    unsigned index;
    unsigned base;
  Bit32u   addr_displacement; // address displacement
  Bit32u   rm_addr;
  Bit32u   Id;
  Bit16u   Iw;
  Bit8u    Ib;
  Bit8u    Ib2; // for ENTER_IwIb
  Bit16u   Iw2; // for JMP_Ap
  unsigned ilen; // instruction length
  unsigned os_32, as_32; // OperandSize/AddressSize is 32bit
  unsigned flags_in, flags_out; // flags needed, flags modified
  void (*ResolveModrmAddr)(BxInstruction_tag *);
  void (*execute)(BxInstruction_tag *);
  void (*rep_execute)(BxInstruction_tag *);
  } BxInstruction_t;

typedef void (*BxExecutePtr_t)(BxInstruction_t *);


#if BX_CPU_LEVEL < 2
  /* no GDTR or IDTR register in an 8086 */
#else
typedef struct {
  Bit32u                 base;      /* base address: 24bits=286,32bits=386 */
  Bit16u                 limit;     /* limit, 16bits */
  } bx_global_segment_reg_t;
#endif




#if BX_USE_TLB
  typedef struct {
    Bit32u lpf; // linear page frame
    Bit32u ppf; // physical page frame
    Bit32u pte_addr; // Page Table Address for updating A & D bits
    Bit32u combined_access;
    } bx_TLB_entry;
#endif  // #if BX_USE_TLB



#ifdef BX_BIG_ENDIAN
typedef struct {
  union {
    Bit32u erx;
    struct {
      Bit16u word_filler;
      union {
        Bit16u rx;
        struct {
          Bit8u rh;
          Bit8u rl;
          } byte;
        };
      } word;
    };
  } bx_gen_reg_t;
#else
typedef struct {
  union {
    Bit32u erx;
    struct {
      union {
        Bit16u rx;
        struct {
          Bit8u rl;
          Bit8u rh;
          } byte;
        };
      Bit16u word_filler;
      } word;
    };
  } bx_gen_reg_t;
#endif



#if BX_USE_SMF == 0
   // normal member functions, use this pointer
#  define BX_CPU_THIS_PTR  this->
#  define BX_SMF
#  define BX_CPU_C_PREFIX  BX_CPU_C::
#else
   // static member functions, use direct access
#  define BX_CPU_THIS_PTR  BX_CPU.
#  define BX_SMF           static
#  define BX_CPU_C_PREFIX
#endif

class BX_CPU_C {

public: // for now...

  // General register set
  // eax: accumulator
  // ebx: base
  // ecx: count
  // edx: data
  // ebp: base pointer
  // esi: source index
  // edi: destination index
  // esp: stack pointer
  bx_gen_reg_t  gen_reg[8];

  /* status and control flags register set */
  bx_flags_reg_t eflags; /* status and control flags */
  Boolean inhibit_interrupts;

  Bit32u eip;    /* instruction pointer */
#if BX_CPU_LEVEL > 0
  /* so that we can back up when handling faults, exceptions, etc.
     we need to store the value of the instruction pointer, before
     each fetch/execute cycle.
   */
  Bit32u prev_eip;
  Bit32u prev_esp;
#endif

  /* user segment register set */
  bx_segment_reg_t  cs; /* code segment register */
  bx_segment_reg_t  ss; /* stack segment register */
  bx_segment_reg_t  ds; /* data segment register */
  bx_segment_reg_t  es; /* extra segment register */
  bx_segment_reg_t  fs; /* extra segment register */
  bx_segment_reg_t  gs; /* extra segment register */

  /* system segment registers */
#if BX_CPU_LEVEL >= 2
  bx_global_segment_reg_t gdtr; /* global descriptor table register */
  bx_global_segment_reg_t idtr; /* interrupt descriptor table register */
#endif
  bx_segment_reg_t        ldtr; /* interrupt descriptor table register */
  bx_segment_reg_t        tr;   /* task register */


  /* debug registers 0-7 (unimplemented) */
#if BX_CPU_LEVEL >= 3
  Bit32u dr0;
  Bit32u dr1;
  Bit32u dr2;
  Bit32u dr3;
  Bit32u dr4;
  Bit32u dr5;
  Bit32u dr6;
  Bit32u dr7;
#endif

  /* TR3 - TR7 (Test Register 3-7), unimplemented */
#if 0
  Bit32u tr3;
  Bit32u tr4;
  Bit32u tr5;
  Bit32u tr6;
  Bit32u tr7;
#endif

  /* Control registers */
#if BX_CPU_LEVEL >= 2
  bx_cr0_t  cr0;
  Bit32u    cr1;
  Bit32u    cr2;
  Bit32u    cr3;
#endif
#if BX_CPU_LEVEL >= 5
  Bit32u    cr4;
#endif

  Boolean EXT; /* 1 if processing external interrupt or exception
                * or if not related to current instruction,
                * 0 if current CS:IP caused exception */
  unsigned errorno;   /* signal exception during instruction emulation */

  Boolean  single_step_event;
  volatile Boolean async_event;
  Bit32u   lf_flags_status;
  Boolean  lf_pf;
  bx_lf_flags_entry oszapc;
  bx_lf_flags_entry oszap;


  // for accessing registers by index number
  Bit16u *_16bit_base_reg[8];
  Bit16u *_16bit_index_reg[8];
  Bit32u empty_register;

  // for decoding instructions; accessing seg reg's by index
  bx_segment_reg_t *sreg_mod00_rm16[8];
  bx_segment_reg_t *sreg_mod01_rm16[8];
  bx_segment_reg_t *sreg_mod10_rm16[8];
  bx_segment_reg_t *sreg_mod00_rm32[8];
  bx_segment_reg_t *sreg_mod01_rm32[8];
  bx_segment_reg_t *sreg_mod10_rm32[8];
  bx_segment_reg_t *sreg_mod0_base32[8];
  bx_segment_reg_t *sreg_mod1or2_base32[8];

  // for exceptions
  jmp_buf jmp_buf_env;
  Bit8u curr_exception[2];

  static const Boolean is_exception_OK[3][3];

  bx_segment_reg_t save_cs;
  bx_segment_reg_t save_ss;
  Bit32u           save_eip;
  Bit32u           save_esp;

  // For prefetch'ing instructions
  Bit32u bytesleft;
  Bit8u *fetch_ptr;
  Bit32u prev_linear_page;
  Bit32u prev_phy_page;
  Bit32u max_phy_addr;

  // for paging
#if BX_USE_TLB
  struct {
    bx_TLB_entry entry[BX_TLB_SIZE];
    } TLB;
#endif

  struct {
    Bit32u  paddress1;  // physical address after translation of 1st len1 bytes of data
    Bit32u  paddress2;  // physical address after translation of 2nd len2 bytes of data
    Bit32u  len1;       // number of bytes in page 1
    Bit32u  len2;       // number of bytes in page 2
    unsigned pages;     // number of pages access spans (1 or 2)
    } address_xlation;

  // for lazy flags processing
  BX_SMF Boolean get_OF(void);
  BX_SMF Boolean get_SF(void);
  BX_SMF Boolean get_ZF(void);
  BX_SMF Boolean get_AF(void);
  BX_SMF Boolean get_PF(void);
  BX_SMF Boolean get_CF(void);

  // constructors & destructors...
  BX_CPU_C(void);
  ~BX_CPU_C(void);

  // prototypes for CPU instructions...
  BX_SMF INLINE void ADD_EbGb(BxInstruction_t *);
  BX_SMF INLINE void ADD_EdGd(BxInstruction_t *);
  BX_SMF INLINE void ADD_GbEb(BxInstruction_t *);
  BX_SMF INLINE void ADD_GdEd(BxInstruction_t *);
  BX_SMF INLINE void ADD_ALIb(BxInstruction_t *);
  BX_SMF INLINE void ADD_EAXId(BxInstruction_t *);
  BX_SMF INLINE void OR_EbGb(BxInstruction_t *);
  BX_SMF INLINE void OR_EdGd(BxInstruction_t *);
  BX_SMF INLINE void OR_EwGw(BxInstruction_t *);
  BX_SMF INLINE void OR_GbEb(BxInstruction_t *);
  BX_SMF INLINE void OR_GdEd(BxInstruction_t *);
  BX_SMF INLINE void OR_GwEw(BxInstruction_t *);
  BX_SMF INLINE void OR_ALIb(BxInstruction_t *);
  BX_SMF INLINE void OR_EAXId(BxInstruction_t *);
  BX_SMF INLINE void OR_AXIw(BxInstruction_t *);

  BX_SMF INLINE void PUSH_CS(BxInstruction_t *);
  BX_SMF INLINE void PUSH_DS(BxInstruction_t *);
  BX_SMF INLINE void POP_DS(BxInstruction_t *);
  BX_SMF INLINE void PUSH_ES(BxInstruction_t *);
  BX_SMF INLINE void POP_ES(BxInstruction_t *);
  BX_SMF INLINE void PUSH_FS(BxInstruction_t *);
  BX_SMF INLINE void POP_FS(BxInstruction_t *);
  BX_SMF INLINE void PUSH_GS(BxInstruction_t *);
  BX_SMF INLINE void POP_GS(BxInstruction_t *);
  BX_SMF INLINE void PUSH_SS(BxInstruction_t *);
  BX_SMF INLINE void POP_SS(BxInstruction_t *);

  BX_SMF INLINE void ADC_EbGb(BxInstruction_t *);
  BX_SMF INLINE void ADC_EdGd(BxInstruction_t *);
  BX_SMF INLINE void ADC_GbEb(BxInstruction_t *);
  BX_SMF INLINE void ADC_GdEd(BxInstruction_t *);
  BX_SMF INLINE void ADC_ALIb(BxInstruction_t *);
  BX_SMF INLINE void ADC_EAXId(BxInstruction_t *);
  BX_SMF INLINE void SBB_EbGb(BxInstruction_t *);
  BX_SMF INLINE void SBB_EdGd(BxInstruction_t *);
  BX_SMF INLINE void SBB_GbEb(BxInstruction_t *);
  BX_SMF INLINE void SBB_GdEd(BxInstruction_t *);
  BX_SMF INLINE void SBB_ALIb(BxInstruction_t *);
  BX_SMF INLINE void SBB_EAXId(BxInstruction_t *);

  BX_SMF INLINE void AND_EbGb(BxInstruction_t *);
  BX_SMF INLINE void AND_EdGd(BxInstruction_t *);
  BX_SMF INLINE void AND_EwGw(BxInstruction_t *);
  BX_SMF INLINE void AND_GbEb(BxInstruction_t *);
  BX_SMF INLINE void AND_GdEd(BxInstruction_t *);
  BX_SMF INLINE void AND_GwEw(BxInstruction_t *);
  BX_SMF INLINE void AND_ALIb(BxInstruction_t *);
  BX_SMF INLINE void AND_EAXId(BxInstruction_t *);
  BX_SMF INLINE void AND_AXIw(BxInstruction_t *);
  BX_SMF INLINE void DAA(BxInstruction_t *);
  BX_SMF INLINE void SUB_EbGb(BxInstruction_t *);
  BX_SMF INLINE void SUB_EdGd(BxInstruction_t *);
  BX_SMF INLINE void SUB_GbEb(BxInstruction_t *);
  BX_SMF INLINE void SUB_GdEd(BxInstruction_t *);
  BX_SMF INLINE void SUB_ALIb(BxInstruction_t *);
  BX_SMF INLINE void SUB_EAXId(BxInstruction_t *);
  BX_SMF INLINE void DAS(BxInstruction_t *);

  BX_SMF INLINE void XOR_EbGb(BxInstruction_t *);
  BX_SMF INLINE void XOR_EdGd(BxInstruction_t *);
  BX_SMF INLINE void XOR_EwGw(BxInstruction_t *);
  BX_SMF INLINE void XOR_GbEb(BxInstruction_t *);
  BX_SMF INLINE void XOR_GdEd(BxInstruction_t *);
  BX_SMF INLINE void XOR_GwEw(BxInstruction_t *);
  BX_SMF INLINE void XOR_ALIb(BxInstruction_t *);
  BX_SMF INLINE void XOR_EAXId(BxInstruction_t *);
  BX_SMF INLINE void XOR_AXIw(BxInstruction_t *);
  BX_SMF INLINE void AAA(BxInstruction_t *);
  BX_SMF INLINE void CMP_EbGb(BxInstruction_t *);
  BX_SMF INLINE void CMP_EdGd(BxInstruction_t *);
  BX_SMF INLINE void CMP_GbEb(BxInstruction_t *);
  BX_SMF INLINE void CMP_GdEd(BxInstruction_t *);
  BX_SMF INLINE void CMP_ALIb(BxInstruction_t *);
  BX_SMF INLINE void CMP_EAXId(BxInstruction_t *);
  BX_SMF INLINE void AAS(BxInstruction_t *);

  BX_SMF INLINE void PUSHAD32(BxInstruction_t *);
  BX_SMF INLINE void PUSHAD16(BxInstruction_t *);
  BX_SMF INLINE void POPAD32(BxInstruction_t *);
  BX_SMF INLINE void POPAD16(BxInstruction_t *);
  BX_SMF INLINE void BOUND_GvMa(BxInstruction_t *);
  BX_SMF INLINE void ARPL_EwGw(BxInstruction_t *);
  BX_SMF INLINE void PUSH_Id(BxInstruction_t *);
  BX_SMF INLINE void PUSH_Iw(BxInstruction_t *);
  BX_SMF INLINE void IMUL_GdEdId(BxInstruction_t *);
  BX_SMF INLINE void INSB_YbDX(BxInstruction_t *);
  BX_SMF INLINE void INSW_YvDX(BxInstruction_t *);
  BX_SMF INLINE void OUTSB_DXXb(BxInstruction_t *);
  BX_SMF INLINE void OUTSW_DXXv(BxInstruction_t *);

  BX_SMF INLINE void TEST_EbGb(BxInstruction_t *);
  BX_SMF INLINE void TEST_EdGd(BxInstruction_t *);
  BX_SMF INLINE void TEST_EwGw(BxInstruction_t *);
  BX_SMF INLINE void XCHG_EbGb(BxInstruction_t *);
  BX_SMF INLINE void XCHG_EdGd(BxInstruction_t *);
  BX_SMF INLINE void XCHG_EwGw(BxInstruction_t *);
  BX_SMF INLINE void MOV_EbGb(BxInstruction_t *);
  BX_SMF INLINE void MOV_EdGd(BxInstruction_t *);
  BX_SMF INLINE void MOV_EwGw(BxInstruction_t *);
  BX_SMF INLINE void MOV_GbEb(BxInstruction_t *);
  BX_SMF INLINE void MOV_GdEd(BxInstruction_t *);
  BX_SMF INLINE void MOV_GwEw(BxInstruction_t *);
  BX_SMF INLINE void MOV_EwSw(BxInstruction_t *);
  BX_SMF INLINE void LEA_GdM(BxInstruction_t *);
  BX_SMF INLINE void LEA_GwM(BxInstruction_t *);
  BX_SMF INLINE void MOV_SwEw(BxInstruction_t *);
  BX_SMF INLINE void POP_Ev(BxInstruction_t *);

  BX_SMF INLINE void CBW(BxInstruction_t *);
  BX_SMF INLINE void CWD(BxInstruction_t *);
  BX_SMF INLINE void CALL32_Ap(BxInstruction_t *);
  BX_SMF INLINE void CALL16_Ap(BxInstruction_t *);
  BX_SMF INLINE void WAIT(BxInstruction_t *);
  BX_SMF INLINE void PUSHF_Fv(BxInstruction_t *);
  BX_SMF INLINE void POPF_Fv(BxInstruction_t *);
  BX_SMF INLINE void SAHF(BxInstruction_t *);
  BX_SMF INLINE void LAHF(BxInstruction_t *);

  BX_SMF INLINE void MOV_ALOb(BxInstruction_t *);
  BX_SMF INLINE void MOV_EAXOd(BxInstruction_t *);
  BX_SMF INLINE void MOV_AXOw(BxInstruction_t *);
  BX_SMF INLINE void MOV_ObAL(BxInstruction_t *);
  BX_SMF INLINE void MOV_OdEAX(BxInstruction_t *);
  BX_SMF INLINE void MOV_OwAX(BxInstruction_t *);
  BX_SMF INLINE void MOVSB_XbYb(BxInstruction_t *);
  BX_SMF INLINE void MOVSW_XvYv(BxInstruction_t *);
  BX_SMF INLINE void CMPSB_XbYb(BxInstruction_t *);
  BX_SMF INLINE void CMPSW_XvYv(BxInstruction_t *);
  BX_SMF INLINE void TEST_ALIb(BxInstruction_t *);
  BX_SMF INLINE void TEST_EAXId(BxInstruction_t *);
  BX_SMF INLINE void TEST_AXIw(BxInstruction_t *);
  BX_SMF INLINE void STOSB_YbAL(BxInstruction_t *);
  BX_SMF INLINE void STOSW_YveAX(BxInstruction_t *);
  BX_SMF INLINE void LODSB_ALXb(BxInstruction_t *);
  BX_SMF INLINE void LODSW_eAXXv(BxInstruction_t *);
  BX_SMF INLINE void SCASB_ALXb(BxInstruction_t *);
  BX_SMF INLINE void SCASW_eAXXv(BxInstruction_t *);

  BX_SMF INLINE void RETnear32(BxInstruction_t *);
  BX_SMF INLINE void RETnear16(BxInstruction_t *);
  BX_SMF INLINE void LES_GvMp(BxInstruction_t *);
  BX_SMF INLINE void LDS_GvMp(BxInstruction_t *);
  BX_SMF INLINE void MOV_EbIb(BxInstruction_t *);
  BX_SMF INLINE void MOV_EdId(BxInstruction_t *);
  BX_SMF INLINE void MOV_EwIw(BxInstruction_t *);
  BX_SMF INLINE void ENTER_IwIb(BxInstruction_t *);
  BX_SMF INLINE void LEAVE(BxInstruction_t *);
  BX_SMF INLINE void RETfar32(BxInstruction_t *);
  BX_SMF INLINE void RETfar16(BxInstruction_t *);

  BX_SMF INLINE void INT3(BxInstruction_t *);
  BX_SMF INLINE void INT_Ib(BxInstruction_t *);
  BX_SMF INLINE void INTO(BxInstruction_t *);
  BX_SMF INLINE void IRET32(BxInstruction_t *);
  BX_SMF INLINE void IRET16(BxInstruction_t *);

  BX_SMF INLINE void AAM(BxInstruction_t *);
  BX_SMF INLINE void AAD(BxInstruction_t *);
  BX_SMF INLINE void SALC(BxInstruction_t *);
  BX_SMF INLINE void XLAT(BxInstruction_t *);

  BX_SMF INLINE void LOOPNE_Jb(BxInstruction_t *);
  BX_SMF INLINE void LOOPE_Jb(BxInstruction_t *);
  BX_SMF INLINE void LOOP_Jb(BxInstruction_t *);
  BX_SMF INLINE void JCXZ_Jb(BxInstruction_t *);
  BX_SMF INLINE void IN_ALIb(BxInstruction_t *);
  BX_SMF INLINE void IN_eAXIb(BxInstruction_t *);
  BX_SMF INLINE void OUT_IbAL(BxInstruction_t *);
  BX_SMF INLINE void OUT_IbeAX(BxInstruction_t *);
  BX_SMF INLINE void CALL_Aw(BxInstruction_t *);
  BX_SMF INLINE void CALL_Ad(BxInstruction_t *);
  BX_SMF INLINE void JMP_Jd(BxInstruction_t *);
  BX_SMF INLINE void JMP_Jw(BxInstruction_t *);
  BX_SMF INLINE void JMP_Ap(BxInstruction_t *);
  BX_SMF INLINE void IN_ALDX(BxInstruction_t *);
  BX_SMF INLINE void IN_eAXDX(BxInstruction_t *);
  BX_SMF INLINE void OUT_DXAL(BxInstruction_t *);
  BX_SMF INLINE void OUT_DXeAX(BxInstruction_t *);

  BX_SMF INLINE void HLT(BxInstruction_t *);
  BX_SMF INLINE void CMC(BxInstruction_t *);
  BX_SMF INLINE void CLC(BxInstruction_t *);
  BX_SMF INLINE void STC(BxInstruction_t *);
  BX_SMF INLINE void CLI(BxInstruction_t *);
  BX_SMF INLINE void STI(BxInstruction_t *);
  BX_SMF INLINE void CLD(BxInstruction_t *);
  BX_SMF INLINE void STD(BxInstruction_t *);


  BX_SMF INLINE void LAR_GvEw(BxInstruction_t *);
  BX_SMF INLINE void LSL_GvEw(BxInstruction_t *);
  BX_SMF INLINE void CLTS(BxInstruction_t *);
  BX_SMF INLINE void INVD(BxInstruction_t *);
  BX_SMF INLINE void WBINVD(BxInstruction_t *);

  BX_SMF INLINE void INVLPG_Ea(BxInstruction_t *);

  BX_SMF INLINE void MOV_CdRd(BxInstruction_t *);
  BX_SMF INLINE void MOV_DdRd(BxInstruction_t *);
  BX_SMF INLINE void MOV_RdCd(BxInstruction_t *);
  BX_SMF INLINE void MOV_RdDd(BxInstruction_t *);
  BX_SMF INLINE void MOV_TdRd(BxInstruction_t *);
  BX_SMF INLINE void MOV_RdTd(BxInstruction_t *);

  BX_SMF void JCC_Jd(BxInstruction_t *);
  BX_SMF void JCC_Jw(BxInstruction_t *);

  BX_SMF INLINE void SETO_Eb(BxInstruction_t *);
  BX_SMF INLINE void SETNO_Eb(BxInstruction_t *);
  BX_SMF INLINE void SETB_Eb(BxInstruction_t *);
  BX_SMF INLINE void SETNB_Eb(BxInstruction_t *);
  BX_SMF INLINE void SETZ_Eb(BxInstruction_t *);
  BX_SMF INLINE void SETNZ_Eb(BxInstruction_t *);
  BX_SMF INLINE void SETBE_Eb(BxInstruction_t *);
  BX_SMF INLINE void SETNBE_Eb(BxInstruction_t *);
  BX_SMF INLINE void SETS_Eb(BxInstruction_t *);
  BX_SMF INLINE void SETNS_Eb(BxInstruction_t *);
  BX_SMF INLINE void SETP_Eb(BxInstruction_t *);
  BX_SMF INLINE void SETNP_Eb(BxInstruction_t *);
  BX_SMF INLINE void SETL_Eb(BxInstruction_t *);
  BX_SMF INLINE void SETNL_Eb(BxInstruction_t *);
  BX_SMF INLINE void SETLE_Eb(BxInstruction_t *);
  BX_SMF INLINE void SETNLE_Eb(BxInstruction_t *);

  BX_SMF INLINE void CPUID(BxInstruction_t *);
  BX_SMF INLINE void BT_EvGv(BxInstruction_t *);
  BX_SMF INLINE void SHLD_EdGd(BxInstruction_t *);
  BX_SMF INLINE void SHLD_EwGw(BxInstruction_t *);


  BX_SMF INLINE void BTS_EvGv(BxInstruction_t *);

  BX_SMF INLINE void SHRD_EwGw(BxInstruction_t *);
  BX_SMF void SHRD_EdGd(BxInstruction_t *);

  BX_SMF INLINE void IMUL_GdEd(BxInstruction_t *);

  BX_SMF INLINE void LSS_GvMp(BxInstruction_t *);
  BX_SMF INLINE void BTR_EvGv(BxInstruction_t *);
  BX_SMF INLINE void LFS_GvMp(BxInstruction_t *);
  BX_SMF INLINE void LGS_GvMp(BxInstruction_t *);
  BX_SMF INLINE void MOVZX_GdEb(BxInstruction_t *);
  BX_SMF INLINE void MOVZX_GwEb(BxInstruction_t *);
  BX_SMF INLINE void MOVZX_GdEw(BxInstruction_t *);
  BX_SMF INLINE void MOVZX_GwEw(BxInstruction_t *);
  BX_SMF INLINE void BTC_EvGv(BxInstruction_t *);
  BX_SMF INLINE void BSF_GvEv(BxInstruction_t *);
  BX_SMF INLINE void BSR_GvEv(BxInstruction_t *);
  BX_SMF INLINE void MOVSX_GdEb(BxInstruction_t *);
  BX_SMF INLINE void MOVSX_GwEb(BxInstruction_t *);
  BX_SMF INLINE void MOVSX_GdEw(BxInstruction_t *);
  BX_SMF INLINE void MOVSX_GwEw(BxInstruction_t *);

  BX_SMF INLINE void BSWAP_EAX(BxInstruction_t *);
  BX_SMF INLINE void BSWAP_ECX(BxInstruction_t *);
  BX_SMF INLINE void BSWAP_EDX(BxInstruction_t *);
  BX_SMF INLINE void BSWAP_EBX(BxInstruction_t *);
  BX_SMF INLINE void BSWAP_ESP(BxInstruction_t *);
  BX_SMF INLINE void BSWAP_EBP(BxInstruction_t *);
  BX_SMF INLINE void BSWAP_ESI(BxInstruction_t *);
  BX_SMF INLINE void BSWAP_EDI(BxInstruction_t *);

  BX_SMF INLINE void ADD_EbIb(BxInstruction_t *);
  BX_SMF INLINE void ADC_EbIb(BxInstruction_t *);
  BX_SMF INLINE void SBB_EbIb(BxInstruction_t *);
  BX_SMF INLINE void SUB_EbIb(BxInstruction_t *);
  BX_SMF INLINE void CMP_EbIb(BxInstruction_t *);

  BX_SMF INLINE void XOR_EbIb(BxInstruction_t *);
  BX_SMF INLINE void OR_EbIb(BxInstruction_t *);
  BX_SMF INLINE void AND_EbIb(BxInstruction_t *);

  BX_SMF INLINE void ADD_EdId(BxInstruction_t *);
  BX_SMF INLINE void OR_EdId(BxInstruction_t *);
  BX_SMF INLINE void OR_EwIw(BxInstruction_t *);
  BX_SMF INLINE void ADC_EdId(BxInstruction_t *);
  BX_SMF INLINE void SBB_EdId(BxInstruction_t *);
  BX_SMF INLINE void AND_EdId(BxInstruction_t *);
  BX_SMF INLINE void AND_EwIw(BxInstruction_t *);
  BX_SMF INLINE void SUB_EdId(BxInstruction_t *);
  BX_SMF INLINE void XOR_EdId(BxInstruction_t *);
  BX_SMF INLINE void XOR_EwIw(BxInstruction_t *);
  BX_SMF INLINE void CMP_EdId(BxInstruction_t *);

  BX_SMF INLINE void ROL_Eb(BxInstruction_t *);
  BX_SMF INLINE void ROR_Eb(BxInstruction_t *);
  BX_SMF INLINE void RCL_Eb(BxInstruction_t *);
  BX_SMF INLINE void RCR_Eb(BxInstruction_t *);
  BX_SMF INLINE void SHL_Eb(BxInstruction_t *);
  BX_SMF INLINE void SHR_Eb(BxInstruction_t *);
  BX_SMF INLINE void SAR_Eb(BxInstruction_t *);

  BX_SMF INLINE void ROL_Ed(BxInstruction_t *);
  BX_SMF INLINE void ROL_Ew(BxInstruction_t *);
  BX_SMF INLINE void ROR_Ed(BxInstruction_t *);
  BX_SMF INLINE void ROR_Ew(BxInstruction_t *);
  BX_SMF INLINE void RCL_Ed(BxInstruction_t *);
  BX_SMF INLINE void RCL_Ew(BxInstruction_t *);
  BX_SMF INLINE void RCR_Ed(BxInstruction_t *);
  BX_SMF INLINE void RCR_Ew(BxInstruction_t *);
  BX_SMF INLINE void SHL_Ed(BxInstruction_t *);
  BX_SMF INLINE void SHL_Ew(BxInstruction_t *);
  BX_SMF INLINE void SHR_Ed(BxInstruction_t *);
  BX_SMF INLINE void SHR_Ew(BxInstruction_t *);
  BX_SMF INLINE void SAR_Ed(BxInstruction_t *);
  BX_SMF INLINE void SAR_Ew(BxInstruction_t *);

  BX_SMF INLINE void TEST_EbIb(BxInstruction_t *);
  BX_SMF INLINE void NOT_Eb(BxInstruction_t *);
  BX_SMF INLINE void NEG_Eb(BxInstruction_t *);
  BX_SMF INLINE void MUL_ALEb(BxInstruction_t *);
  BX_SMF INLINE void IMUL_ALEb(BxInstruction_t *);
  BX_SMF INLINE void DIV_ALEb(BxInstruction_t *);
  BX_SMF INLINE void IDIV_ALEb(BxInstruction_t *);

  BX_SMF INLINE void TEST_EdId(BxInstruction_t *);
  BX_SMF INLINE void TEST_EwIw(BxInstruction_t *);
  BX_SMF INLINE void NOT_Ed(BxInstruction_t *);
  BX_SMF INLINE void NOT_Ew(BxInstruction_t *);
  BX_SMF INLINE void NEG_Ed(BxInstruction_t *);
  BX_SMF INLINE void MUL_EAXEd(BxInstruction_t *);
  BX_SMF INLINE void IMUL_EAXEd(BxInstruction_t *);
  BX_SMF INLINE void DIV_EAXEd(BxInstruction_t *);
  BX_SMF INLINE void IDIV_EAXEd(BxInstruction_t *);

  BX_SMF INLINE void INC_Eb(BxInstruction_t *);
  BX_SMF INLINE void DEC_Eb(BxInstruction_t *);

  BX_SMF INLINE void INC_Ed(BxInstruction_t *);
  BX_SMF INLINE void DEC_Ed(BxInstruction_t *);
  BX_SMF INLINE void CALL_Ed(BxInstruction_t *);
  BX_SMF INLINE void CALL_Ew(BxInstruction_t *);
  BX_SMF INLINE void CALL32_Ep(BxInstruction_t *);
  BX_SMF INLINE void CALL16_Ep(BxInstruction_t *);
  BX_SMF INLINE void JMP_Ed(BxInstruction_t *);
  BX_SMF INLINE void JMP_Ew(BxInstruction_t *);
  BX_SMF INLINE void JMP32_Ep(BxInstruction_t *);
  BX_SMF INLINE void JMP16_Ep(BxInstruction_t *);
  BX_SMF INLINE void PUSH_Ed(BxInstruction_t *);
  BX_SMF INLINE void PUSH_Ew(BxInstruction_t *);

  BX_SMF INLINE void SLDT_Ew(BxInstruction_t *);
  BX_SMF INLINE void STR_Ew(BxInstruction_t *);
  BX_SMF INLINE void LLDT_Ew(BxInstruction_t *);
  BX_SMF INLINE void LTR_Ew(BxInstruction_t *);
  BX_SMF INLINE void VERR_Ew(BxInstruction_t *);
  BX_SMF INLINE void VERW_Ew(BxInstruction_t *);

  BX_SMF INLINE void SGDT_Ms(BxInstruction_t *);
  BX_SMF INLINE void SIDT_Ms(BxInstruction_t *);
  BX_SMF INLINE void LGDT_Ms(BxInstruction_t *);
  BX_SMF INLINE void LIDT_Ms(BxInstruction_t *);
  BX_SMF INLINE void SMSW_Ew(BxInstruction_t *);
  BX_SMF INLINE void LMSW_Ew(BxInstruction_t *);


  BX_SMF INLINE void BT_EvIb(BxInstruction_t *);
  BX_SMF INLINE void BTS_EvIb(BxInstruction_t *);
  BX_SMF INLINE void BTR_EvIb(BxInstruction_t *);
  BX_SMF INLINE void BTC_EvIb(BxInstruction_t *);

  BX_SMF void ESC0(BxInstruction_t *);
  BX_SMF void ESC1(BxInstruction_t *);
  BX_SMF void ESC2(BxInstruction_t *);
  BX_SMF void ESC3(BxInstruction_t *);
  BX_SMF void ESC4(BxInstruction_t *);
  BX_SMF void ESC5(BxInstruction_t *);
  BX_SMF void ESC6(BxInstruction_t *);
  BX_SMF void ESC7(BxInstruction_t *);

  BX_SMF INLINE void FADD_Es(BxInstruction_t *);
  BX_SMF INLINE void FMUL_Es(BxInstruction_t *);
  BX_SMF INLINE void FCOM_Es(BxInstruction_t *);
  BX_SMF INLINE void FCOMP_Es(BxInstruction_t *);
  BX_SMF INLINE void FSUB_Es(BxInstruction_t *);
  BX_SMF INLINE void FSUBR_Es(BxInstruction_t *);
  BX_SMF INLINE void FDIV_Es(BxInstruction_t *);
  BX_SMF INLINE void FDIVR_Es(BxInstruction_t *);

  BX_SMF INLINE void FADD_ST_STi(BxInstruction_t *);
  BX_SMF INLINE void FMUL_ST_STi(BxInstruction_t *);
  BX_SMF INLINE void FCOM_ST_STi(BxInstruction_t *);
  BX_SMF INLINE void FCOMP_ST_STi(BxInstruction_t *);
  BX_SMF INLINE void FSUB_ST_STi(BxInstruction_t *);
  BX_SMF INLINE void FSUBR_ST_STi(BxInstruction_t *);
  BX_SMF INLINE void FDIV_ST_STi(BxInstruction_t *);
  BX_SMF INLINE void FDIVR_ST_STi(BxInstruction_t *);

  BX_SMF INLINE void FLD_Es(BxInstruction_t *);
  BX_SMF INLINE void FST_Es(BxInstruction_t *);
  BX_SMF INLINE void FSTP_Es(BxInstruction_t *);
  BX_SMF INLINE void FLDENV_Ea(BxInstruction_t *);
  BX_SMF INLINE void FLDCW_Ew(BxInstruction_t *);
  BX_SMF INLINE void FSTENV_Ea(BxInstruction_t *);
  BX_SMF INLINE void FSTCW_Ew(BxInstruction_t *);

  BX_SMF INLINE void FNOP(BxInstruction_t *);
  BX_SMF INLINE void FCHS(BxInstruction_t *);
  BX_SMF INLINE void FLD1(BxInstruction_t *);
  BX_SMF INLINE void F2XM1(BxInstruction_t *);
  BX_SMF INLINE void FPREM(BxInstruction_t *);

  BX_SMF INLINE void FABS(BxInstruction_t *);
  BX_SMF INLINE void FLDL2T(BxInstruction_t *);
  BX_SMF INLINE void FYL2X(BxInstruction_t *);
  BX_SMF INLINE void FYL2XP1(BxInstruction_t *);

  BX_SMF INLINE void FLDL2E(BxInstruction_t *);
  BX_SMF INLINE void FPTAN(BxInstruction_t *);
  BX_SMF INLINE void FSQRT(BxInstruction_t *);

  BX_SMF INLINE void FXCH_3(BxInstruction_t *);
  BX_SMF INLINE void FLDPI(BxInstruction_t *);
  BX_SMF INLINE void FPATAN(BxInstruction_t *);
  BX_SMF INLINE void FSINCOS(BxInstruction_t *);

  BX_SMF INLINE void FTST(BxInstruction_t *);
  BX_SMF INLINE void FLDLG2(BxInstruction_t *);
  BX_SMF INLINE void FXTRACT(BxInstruction_t *);
  BX_SMF INLINE void FRNDINT(BxInstruction_t *);

  BX_SMF INLINE void FXAM(BxInstruction_t *);
  BX_SMF INLINE void FSCALE(BxInstruction_t *);

  BX_SMF INLINE void FLDZ(BxInstruction_t *);
  BX_SMF INLINE void FDECSTP(BxInstruction_t *);
  BX_SMF INLINE void FSIN(BxInstruction_t *);
  BX_SMF INLINE void FINCSTP(BxInstruction_t *);
  BX_SMF INLINE void FCOS(BxInstruction_t *);

  BX_SMF INLINE void FIADD_Ew(BxInstruction_t *);
  BX_SMF INLINE void FIMUL_Ew(BxInstruction_t *);
  BX_SMF INLINE void FICOM_Ew(BxInstruction_t *);
  BX_SMF INLINE void FICOMP_Ew(BxInstruction_t *);
  BX_SMF INLINE void FISUB_Ew(BxInstruction_t *);
  BX_SMF INLINE void FISUBR_Ew(BxInstruction_t *);
  BX_SMF INLINE void FIDIV_Ew(BxInstruction_t *);
  BX_SMF INLINE void FIDIVR_Ew(BxInstruction_t *);
  BX_SMF INLINE void FCMOVB_ST_STi(BxInstruction_t *);
  BX_SMF INLINE void FCMOVE_ST_STi(BxInstruction_t *);
  BX_SMF INLINE void FCMOVBE_ST_STi(BxInstruction_t *);
  BX_SMF INLINE void FCMOVU_ST_STi(BxInstruction_t *);

  BX_SMF INLINE void FUCOMPP(BxInstruction_t *);

  BX_SMF INLINE void FILD_Ew(BxInstruction_t *);
  BX_SMF INLINE void FIST_Ew(BxInstruction_t *);
  BX_SMF INLINE void FISTP_Ew(BxInstruction_t *);
  BX_SMF INLINE void FLD_Et(BxInstruction_t *);
  BX_SMF INLINE void FSTP_Et(BxInstruction_t *);

  BX_SMF INLINE void FENI(BxInstruction_t *);
  BX_SMF INLINE void FDISI(BxInstruction_t *);
  BX_SMF INLINE void FCLEX(BxInstruction_t *);
  BX_SMF INLINE void FINIT(BxInstruction_t *);
  BX_SMF INLINE void FSETPM(BxInstruction_t *);

  BX_SMF INLINE void FCMOVNB_ST_STi(BxInstruction_t *);
  BX_SMF INLINE void FCMOVNE_ST_STi(BxInstruction_t *);
  BX_SMF INLINE void FCMOVNBE_ST_STi(BxInstruction_t *);
  BX_SMF INLINE void FCMOVNU_ST_STi(BxInstruction_t *);
  BX_SMF INLINE void FUCOMI_ST_STi(BxInstruction_t *);
  BX_SMF INLINE void FCOMI_ST_STi(BxInstruction_t *);

  BX_SMF INLINE void FADD_El(BxInstruction_t *);
  BX_SMF INLINE void FMUL_El(BxInstruction_t *);
  BX_SMF INLINE void FCOM_El(BxInstruction_t *);
  BX_SMF INLINE void FCOMP_El(BxInstruction_t *);
  BX_SMF INLINE void FSUB_El(BxInstruction_t *);
  BX_SMF INLINE void FSUBR_El(BxInstruction_t *);
  BX_SMF INLINE void FDIV_El(BxInstruction_t *);
  BX_SMF INLINE void FDIVR_El(BxInstruction_t *);

  BX_SMF INLINE void FADD_STi_ST(BxInstruction_t *);
  BX_SMF INLINE void FMUL_STi_ST(BxInstruction_t *);
  BX_SMF INLINE void FCOM_STi_ST(BxInstruction_t *);
  BX_SMF INLINE void FCOMP_STi_ST(BxInstruction_t *);
  BX_SMF INLINE void FSUB_STi_ST(BxInstruction_t *);
  BX_SMF INLINE void FSUBR_STi_ST(BxInstruction_t *);
  BX_SMF INLINE void FDIV_STi_ST(BxInstruction_t *);
  BX_SMF INLINE void FDIVR_STi_ST(BxInstruction_t *);

  BX_SMF INLINE void FLD_El(BxInstruction_t *);
  BX_SMF INLINE void FST_El(BxInstruction_t *);
  BX_SMF INLINE void FSTP_El(BxInstruction_t *);
  BX_SMF INLINE void FRSTOR_Ea(BxInstruction_t *);
  BX_SMF INLINE void FSAVE_Ea(BxInstruction_t *);
  BX_SMF INLINE void FSTSW_Ew(BxInstruction_t *);

  BX_SMF INLINE void FFREE_STi(BxInstruction_t *);
  BX_SMF INLINE void FST_STi(BxInstruction_t *);
  BX_SMF INLINE void FSTP_STi(BxInstruction_t *);
  BX_SMF INLINE void FUCOMP_STi(BxInstruction_t *);
  BX_SMF INLINE void FIADD_Ed(BxInstruction_t *);
  BX_SMF INLINE void FIMUL_Ed(BxInstruction_t *);
  BX_SMF INLINE void FICOM_Ed(BxInstruction_t *);
  BX_SMF INLINE void FICOMP_Ed(BxInstruction_t *);
  BX_SMF INLINE void FISUB_Ed(BxInstruction_t *);
  BX_SMF INLINE void FISUBR_Ed(BxInstruction_t *);
  BX_SMF INLINE void FIDIV_Ed(BxInstruction_t *);
  BX_SMF INLINE void FIDIVR_Ed(BxInstruction_t *);

  BX_SMF INLINE void FCOMIP(BxInstruction_t *);
  BX_SMF INLINE void FUCOMIP(BxInstruction_t *);

  BX_SMF INLINE void FADDP_STi_ST(BxInstruction_t *);
  BX_SMF INLINE void FMULP_STi_ST(BxInstruction_t *);
  BX_SMF INLINE void FCOMPP(BxInstruction_t *);
  BX_SMF INLINE void FSUBP_STi_ST(BxInstruction_t *);
  BX_SMF INLINE void FSUBRP_STi_ST(BxInstruction_t *);
  BX_SMF INLINE void FDIVP_STi_ST(BxInstruction_t *);
  BX_SMF INLINE void FDIVRP_STi_ST(BxInstruction_t *);

  BX_SMF INLINE void FILD_Ed(BxInstruction_t *);
  BX_SMF INLINE void FIST_Ed(BxInstruction_t *);
  BX_SMF INLINE void FISTP_Ed(BxInstruction_t *);
  BX_SMF INLINE void FBLD_Eb(BxInstruction_t *);
  BX_SMF INLINE void FILD_Eq(BxInstruction_t *);
  BX_SMF INLINE void FBSTP_Eb(BxInstruction_t *);
  BX_SMF INLINE void FISTP_Eq(BxInstruction_t *);

  BX_SMF INLINE void FSTSW_AX(BxInstruction_t *);

  BX_SMF INLINE void CMPXCHG_XBTS(BxInstruction_t *);
  BX_SMF INLINE void CMPXCHG_IBTS(BxInstruction_t *);
  BX_SMF INLINE void CMPXCHG_EbGb(BxInstruction_t *);
  BX_SMF INLINE void CMPXCHG_EdGd(BxInstruction_t *);
  BX_SMF INLINE void XADD_EbGb(BxInstruction_t *);
  BX_SMF INLINE void XADD_EdGd(BxInstruction_t *);
  BX_SMF INLINE void RETnear32_Iw(BxInstruction_t *);
  BX_SMF INLINE void RETnear16_Iw(BxInstruction_t *);
  BX_SMF INLINE void RETfar32_Iw(BxInstruction_t *);
  BX_SMF INLINE void RETfar16_Iw(BxInstruction_t *);
  BX_SMF INLINE void FLD_STi(BxInstruction_t *);
  BX_SMF INLINE void FXCH_STi(BxInstruction_t *);
  BX_SMF INLINE void FLDLN2(BxInstruction_t *);
  BX_SMF INLINE void FPREM1(BxInstruction_t *);
  BX_SMF INLINE void FUCOM_STi_ST(BxInstruction_t *);

  BX_SMF INLINE void LOADALL(BxInstruction_t *);
  BX_SMF void CMOV_GdEd(BxInstruction_t *);
  BX_SMF void CMOV_GwEw(BxInstruction_t *);

  BX_SMF INLINE void ADD_EwGw(BxInstruction_t *);
  BX_SMF INLINE void ADD_GwEw(BxInstruction_t *);
  BX_SMF INLINE void ADD_AXIw(BxInstruction_t *);
  BX_SMF INLINE void ADC_EwGw(BxInstruction_t *);
  BX_SMF INLINE void ADC_GwEw(BxInstruction_t *);
  BX_SMF INLINE void ADC_AXIw(BxInstruction_t *);
  BX_SMF INLINE void SBB_EwGw(BxInstruction_t *);
  BX_SMF INLINE void SBB_GwEw(BxInstruction_t *);
  BX_SMF INLINE void SBB_AXIw(BxInstruction_t *);
  BX_SMF INLINE void SBB_EwIw(BxInstruction_t *);
  BX_SMF INLINE void SUB_EwGw(BxInstruction_t *);
  BX_SMF INLINE void SUB_GwEw(BxInstruction_t *);
  BX_SMF INLINE void SUB_AXIw(BxInstruction_t *);
  BX_SMF INLINE void CMP_EwGw(BxInstruction_t *);
  BX_SMF INLINE void CMP_GwEw(BxInstruction_t *);
  BX_SMF INLINE void CMP_AXIw(BxInstruction_t *);
  BX_SMF INLINE void CWDE(BxInstruction_t *);
  BX_SMF INLINE void CDQ(BxInstruction_t *);
  BX_SMF INLINE void XADD_EwGw(BxInstruction_t *);
  BX_SMF INLINE void ADD_EwIw(BxInstruction_t *);
  BX_SMF INLINE void ADC_EwIw(BxInstruction_t *);
  BX_SMF INLINE void SUB_EwIw(BxInstruction_t *);
  BX_SMF INLINE void CMP_EwIw(BxInstruction_t *);
  BX_SMF INLINE void NEG_Ew(BxInstruction_t *);
  BX_SMF INLINE void INC_Ew(BxInstruction_t *);
  BX_SMF INLINE void DEC_Ew(BxInstruction_t *);
  BX_SMF INLINE void CMPXCHG_EwGw(BxInstruction_t *);
  BX_SMF INLINE void MUL_AXEw(BxInstruction_t *);
  BX_SMF INLINE void IMUL_AXEw(BxInstruction_t *);
  BX_SMF INLINE void DIV_AXEw(BxInstruction_t *);
  BX_SMF INLINE void IDIV_AXEw(BxInstruction_t *);
  BX_SMF INLINE void IMUL_GwEwIw(BxInstruction_t *);
  BX_SMF INLINE void IMUL_GwEw(BxInstruction_t *);
  BX_SMF void NOP(BxInstruction_t *);
  BX_SMF void MOV_RLIb(BxInstruction_t *);
  BX_SMF void MOV_RHIb(BxInstruction_t *);
  BX_SMF void MOV_RXIw(BxInstruction_t *);
  BX_SMF void MOV_ERXId(BxInstruction_t *);
  BX_SMF void INC_RX(BxInstruction_t *);
  BX_SMF void DEC_RX(BxInstruction_t *);
  BX_SMF void INC_ERX(BxInstruction_t *);
  BX_SMF void DEC_ERX(BxInstruction_t *);
  BX_SMF void PUSH_RX(BxInstruction_t *);
  BX_SMF void POP_RX(BxInstruction_t *);
  BX_SMF void PUSH_ERX(BxInstruction_t *);
  BX_SMF void POP_ERX(BxInstruction_t *);
  BX_SMF void POP_Ew(BxInstruction_t *);
  BX_SMF void POP_Ed(BxInstruction_t *);
  BX_SMF void XCHG_RXAX(BxInstruction_t *);
  BX_SMF void XCHG_ERXEAX(BxInstruction_t *);

  BX_SMF void SetCR0(Bit32u val_32);
  BX_SMF void dynamic_translate(void);
  BX_SMF void dynamic_init(void);
  BX_SMF unsigned FetchDecode(Bit8u *, BxInstruction_t *, unsigned, Boolean);
  BX_SMF void UndefinedOpcode(BxInstruction_t *);

  BX_SMF void Resolve16Mod0Rm0(BxInstruction_t *);
  BX_SMF void Resolve16Mod0Rm1(BxInstruction_t *);
  BX_SMF void Resolve16Mod0Rm2(BxInstruction_t *);
  BX_SMF void Resolve16Mod0Rm3(BxInstruction_t *);
  BX_SMF void Resolve16Mod0Rm4(BxInstruction_t *);
  BX_SMF void Resolve16Mod0Rm5(BxInstruction_t *);
  BX_SMF void Resolve16Mod0Rm7(BxInstruction_t *);

  BX_SMF void Resolve16Mod1or2Rm0(BxInstruction_t *);
  BX_SMF void Resolve16Mod1or2Rm1(BxInstruction_t *);
  BX_SMF void Resolve16Mod1or2Rm2(BxInstruction_t *);
  BX_SMF void Resolve16Mod1or2Rm3(BxInstruction_t *);
  BX_SMF void Resolve16Mod1or2Rm4(BxInstruction_t *);
  BX_SMF void Resolve16Mod1or2Rm5(BxInstruction_t *);
  BX_SMF void Resolve16Mod1or2Rm6(BxInstruction_t *);
  BX_SMF void Resolve16Mod1or2Rm7(BxInstruction_t *);

  BX_SMF void Resolve32Mod0Rm0(BxInstruction_t *);
  BX_SMF void Resolve32Mod0Rm1(BxInstruction_t *);
  BX_SMF void Resolve32Mod0Rm2(BxInstruction_t *);
  BX_SMF void Resolve32Mod0Rm3(BxInstruction_t *);
  BX_SMF void Resolve32Mod0Rm6(BxInstruction_t *);
  BX_SMF void Resolve32Mod0Rm7(BxInstruction_t *);

  BX_SMF void Resolve32Mod1or2Rm0(BxInstruction_t *);
  BX_SMF void Resolve32Mod1or2Rm1(BxInstruction_t *);
  BX_SMF void Resolve32Mod1or2Rm2(BxInstruction_t *);
  BX_SMF void Resolve32Mod1or2Rm3(BxInstruction_t *);
  BX_SMF void Resolve32Mod1or2Rm5(BxInstruction_t *);
  BX_SMF void Resolve32Mod1or2Rm6(BxInstruction_t *);
  BX_SMF void Resolve32Mod1or2Rm7(BxInstruction_t *);

  BX_SMF void Resolve32Mod0Base0(BxInstruction_t *);
  BX_SMF void Resolve32Mod0Base1(BxInstruction_t *);
  BX_SMF void Resolve32Mod0Base2(BxInstruction_t *);
  BX_SMF void Resolve32Mod0Base3(BxInstruction_t *);
  BX_SMF void Resolve32Mod0Base4(BxInstruction_t *);
  BX_SMF void Resolve32Mod0Base5(BxInstruction_t *);
  BX_SMF void Resolve32Mod0Base6(BxInstruction_t *);
  BX_SMF void Resolve32Mod0Base7(BxInstruction_t *);

  BX_SMF void Resolve32Mod1or2Base0(BxInstruction_t *);
  BX_SMF void Resolve32Mod1or2Base1(BxInstruction_t *);
  BX_SMF void Resolve32Mod1or2Base2(BxInstruction_t *);
  BX_SMF void Resolve32Mod1or2Base3(BxInstruction_t *);
  BX_SMF void Resolve32Mod1or2Base4(BxInstruction_t *);
  BX_SMF void Resolve32Mod1or2Base5(BxInstruction_t *);
  BX_SMF void Resolve32Mod1or2Base6(BxInstruction_t *);
  BX_SMF void Resolve32Mod1or2Base7(BxInstruction_t *);


  BX_SMF void REP(void (*)(void));
  BX_SMF void REP_ZF(void (*)(void), unsigned rep_prefix);
#if BX_DEBUGGER
  BX_SMF void     dbg_take_irq(void);
  BX_SMF void     dbg_force_interrupt(unsigned vector);
  BX_SMF void     dbg_take_dma(void);
  BX_SMF Boolean  dbg_get_cpu(bx_dbg_cpu_t *cpu);
  BX_SMF Boolean  dbg_set_cpu(bx_dbg_cpu_t *cpu);
  BX_SMF Boolean  dbg_set_reg(unsigned reg, Bit32u val);
  BX_SMF Bit32u   dbg_get_reg(unsigned reg);
  BX_SMF unsigned dbg_query_pending(void);
  BX_SMF Bit32u   dbg_get_descriptor_l(bx_descriptor_t *);
  BX_SMF Bit32u   dbg_get_descriptor_h(bx_descriptor_t *);
  BX_SMF Bit32u   dbg_get_eflags(void);
  BX_SMF Boolean  dbg_is_begin_instr_bpoint(Bit32u cs, Bit32u eip, Bit32u laddr,
                                            Bit32u is_32);
  BX_SMF Boolean  dbg_is_end_instr_bpoint(Bit32u cs, Bit32u eip,
                                          Bit32u laddr, Bit32u is_32);
#endif
#if BX_DEBUGGER || BX_DISASM || BX_INSTRUMENTATION
  BX_SMF void     dbg_xlate_linear2phy(Bit32u linear, Bit32u *phy, Boolean *valid);
#endif
  BX_SMF void     atexit(void);

  // now for some ancillary functions...
  BX_SMF void cpu_loop(void);
  BX_SMF void decode_exgx16(unsigned need_fetch);
  BX_SMF void decode_exgx32(unsigned need_fetch);

  BX_SMF void prefetch(void);
  BX_SMF void revalidate_prefetch_q(void);
  BX_SMF INLINE void invalidate_prefetch_q(void);

  BX_SMF INLINE Bit8u fetch_next_byte(void);
  BX_SMF INLINE Bit16u fetch_next_word(void);
  BX_SMF INLINE Bit32u fetch_next_dword(void);

  BX_SMF inline void write_virtual_checks(bx_segment_reg_t *seg, Bit32u offset, unsigned length);
  BX_SMF inline void read_virtual_checks(bx_segment_reg_t *seg, Bit32u offset, unsigned length);
  BX_SMF void write_virtual_byte(bx_segment_reg_t *seg, Bit32u offset, Bit8u *data);
  BX_SMF void write_virtual_word(bx_segment_reg_t *seg, Bit32u offset, Bit16u *data);
  BX_SMF void write_virtual_dword(bx_segment_reg_t *seg, Bit32u offset, Bit32u *data);
  BX_SMF void read_virtual_byte(bx_segment_reg_t *seg, Bit32u offset, Bit8u *data);
  BX_SMF void read_virtual_word(bx_segment_reg_t *seg, Bit32u offset, Bit16u *data);
  BX_SMF void read_virtual_dword(bx_segment_reg_t *seg, Bit32u offset, Bit32u *data);

  BX_SMF void read_RMW_virtual_byte(bx_segment_reg_t *seg, Bit32u offset, Bit8u *data);
  BX_SMF void read_RMW_virtual_word(bx_segment_reg_t *seg, Bit32u offset, Bit16u *data);
  BX_SMF void read_RMW_virtual_dword(bx_segment_reg_t *seg, Bit32u offset, Bit32u *data);
  BX_SMF void write_RMW_virtual_byte(Bit8u val8);
  BX_SMF void write_RMW_virtual_word(Bit16u val16);
  BX_SMF void write_RMW_virtual_dword(Bit32u val32);

  BX_SMF void access_linear(Bit32u address, unsigned length, unsigned pl,
                     unsigned rw, void *data);
  BX_SMF Bit32u itranslate_linear(Bit32u laddress, unsigned pl);
  BX_SMF Bit32u dtranslate_linear(Bit32u laddress, unsigned pl, unsigned rw);
  BX_SMF void TLB_flush(void);
  BX_SMF void TLB_init(void);
  BX_SMF void set_INTR(Boolean value);
  BX_SMF char *strseg(bx_segment_reg_t *seg);
  BX_SMF void interrupt(Bit8u vector, Boolean is_INT, Boolean is_error_code,
                 Bit16u error_code);
#if BX_CPU_LEVEL >= 2
  BX_SMF void exception(unsigned vector, Bit16u error_code, Boolean is_INT);
#endif
  BX_SMF int  int_number(bx_segment_reg_t *seg);
  BX_SMF void shutdown_cpu(void);
  BX_SMF void enable_paging(void);
  BX_SMF void disable_paging(void);
  BX_SMF void CR3_change(Bit32u value32);
  BX_SMF void reset(unsigned source);

  BX_SMF void jump_protected(BxInstruction_t *, Bit16u cs, Bit32u disp32);
  BX_SMF void call_protected(BxInstruction_t *, Bit16u cs, Bit32u disp32);
  BX_SMF void return_protected(BxInstruction_t *, Bit16u pop_bytes);
  BX_SMF void iret_protected(BxInstruction_t *);
  BX_SMF void validate_seg_regs(void);
  BX_SMF void stack_return_to_v86(Bit32u new_eip, Bit32u raw_cs_selector,
                                     Bit32u flags32);
  BX_SMF void stack_return_from_v86(BxInstruction_t *);
  BX_SMF void init_v8086_mode(void);
  BX_SMF void v8086_message(void);
  BX_SMF void task_switch(bx_selector_t *selector,
                     bx_descriptor_t *descriptor,
                     unsigned source,
                     Bit32u dword1, Bit32u dword2);
  BX_SMF void write_flags(Bit16u flags, Boolean change_IOPL, Boolean change_IF);
  BX_SMF void write_eflags(Bit32u eflags, Boolean change_IOPL, Boolean change_IF,
                    Boolean change_VM, Boolean change_RF);
  BX_SMF Bit16u read_flags(void);
  BX_SMF Bit32u read_eflags(void);

  BX_SMF Bit8u   inp8(Bit16u addr);
  BX_SMF void    outp8(Bit16u addr, Bit8u value);
  BX_SMF Bit16u  inp16(Bit16u addr);
  BX_SMF void    outp16(Bit16u addr, Bit16u value);
  BX_SMF Bit32u  inp32(Bit16u addr);
  BX_SMF void    outp32(Bit16u addr, Bit32u value);
  BX_SMF Boolean allow_io(Bit16u addr, unsigned len);
  BX_SMF void    enter_protected_mode(void);
  BX_SMF void    enter_real_mode(void);
  BX_SMF Bit16u  read_seg_reg(Bit32u index);
  BX_SMF void    write_seg_reg(Bit32u index, Bit16u value);
  BX_SMF void    parse_selector(Bit16u raw_selector, bx_selector_t *selector);
  BX_SMF void    parse_descriptor(Bit32u dword1, Bit32u dword2, bx_descriptor_t *temp);
  BX_SMF void    load_ldtr(bx_selector_t *selector, bx_descriptor_t *descriptor);
  BX_SMF void    load_cs(bx_selector_t *selector, bx_descriptor_t *descriptor, Bit8u cpl);
  BX_SMF void    load_ss(bx_selector_t *selector, bx_descriptor_t *descriptor, Bit8u cpl);
  BX_SMF void    fetch_raw_descriptor(bx_selector_t *selector,
                               Bit32u *dword1, Bit32u *dword2, Bit8u exception);
  BX_SMF void    load_seg_reg(bx_segment_reg_t *seg, Bit16u new_value);
  BX_SMF Boolean fetch_raw_descriptor2(bx_selector_t *selector,
                                Bit32u *dword1, Bit32u *dword2);
  BX_SMF void    push_16(Bit16u value16);
  BX_SMF void    push_32(Bit32u value32);
  BX_SMF void    pop_16(Bit16u *value16_ptr);
  BX_SMF void    pop_32(Bit32u *value32_ptr);
  BX_SMF Boolean can_push(bx_descriptor_t *descriptor, Bit32u esp, Bit32u bytes);
  BX_SMF Boolean can_pop(Bit32u bytes);
  BX_SMF void    sanity_checks(void);

  BX_SMF void    debug(Bit32u offset);

  BX_SMF inline void set_CF(Boolean val);
  BX_SMF inline void set_AF(Boolean val);
  BX_SMF inline void set_ZF(Boolean val);
  BX_SMF inline void set_SF(Boolean val);
  BX_SMF inline void set_OF(Boolean val);
  BX_SMF inline void set_PF(Boolean val);
  BX_SMF inline void set_PF_base(Bit8u val);


  BX_SMF inline void set_AX(Bit16u ax);
  BX_SMF inline void set_BX(Bit16u bx);
  BX_SMF inline void set_CX(Bit16u cx);
  BX_SMF inline void set_DX(Bit16u dx);
  BX_SMF inline void set_AL(Bit8u  al);
  BX_SMF inline void set_AH(Bit8u  ah);
  BX_SMF inline void set_BL(Bit8u  bl);
  BX_SMF inline void set_BH(Bit8u  bh);
  BX_SMF inline void set_CL(Bit8u  cl);
  BX_SMF inline void set_CH(Bit8u  ch);
  BX_SMF inline void set_DL(Bit8u  dl);
  BX_SMF inline void set_DH(Bit8u  dh);

  BX_SMF inline Bit8u get_AL(void);
  BX_SMF inline Bit8u get_AH(void);
  BX_SMF inline Bit8u get_BL(void);
  BX_SMF inline Bit8u get_BH(void);
  BX_SMF inline Bit8u get_CL(void);
  BX_SMF inline Bit8u get_CH(void);
  BX_SMF inline Bit8u get_DL(void);
  BX_SMF inline Bit8u get_DH(void);

  BX_SMF inline Bit16u get_AX(void);
  BX_SMF inline Bit16u get_BX(void);
  BX_SMF inline Bit16u get_CX(void);
  BX_SMF inline Bit16u get_DX(void);

#if BX_CPU_LEVEL >= 2
  BX_SMF inline Boolean real_mode(void);
#endif
#if BX_CPU_LEVEL >= 3
  BX_SMF inline Boolean protected_mode(void);
  BX_SMF inline Boolean v8086_mode(void);
#endif
  };



extern BX_CPU_C       BX_CPU;


BX_SMF inline void BX_CPU_C_PREFIX set_AX(Bit16u ax) { AX = ax; };
BX_SMF inline void BX_CPU_C_PREFIX set_BX(Bit16u bx) { BX = bx; };
BX_SMF inline void BX_CPU_C_PREFIX set_CX(Bit16u cx) { CX = cx; };
BX_SMF inline void BX_CPU_C_PREFIX set_DX(Bit16u dx) { DX = dx; };
BX_SMF inline void BX_CPU_C_PREFIX set_AL(Bit8u  al) { AL = al; };
BX_SMF inline void BX_CPU_C_PREFIX set_AH(Bit8u  ah) { AH = ah; };
BX_SMF inline void BX_CPU_C_PREFIX set_BL(Bit8u  bl) { BL = bl; };
BX_SMF inline void BX_CPU_C_PREFIX set_BH(Bit8u  bh) { BH = bh; };
BX_SMF inline void BX_CPU_C_PREFIX set_CL(Bit8u  cl) { CL = cl; };
BX_SMF inline void BX_CPU_C_PREFIX set_CH(Bit8u  ch) { CH = ch; };
BX_SMF inline void BX_CPU_C_PREFIX set_DL(Bit8u  dl) { DL = dl; };
BX_SMF inline void BX_CPU_C_PREFIX set_DH(Bit8u  dh) { DH = dh; };

BX_SMF inline Bit8u BX_CPU_C_PREFIX get_AL(void) { return(AL); };
BX_SMF inline Bit8u BX_CPU_C_PREFIX get_AH(void) { return(AH); };
BX_SMF inline Bit8u BX_CPU_C_PREFIX get_BL(void) { return(BL); };
BX_SMF inline Bit8u BX_CPU_C_PREFIX get_BH(void) { return(BH); };
BX_SMF inline Bit8u BX_CPU_C_PREFIX get_CL(void) { return(CL); };
BX_SMF inline Bit8u BX_CPU_C_PREFIX get_CH(void) { return(CH); };
BX_SMF inline Bit8u BX_CPU_C_PREFIX get_DL(void) { return(DL); };
BX_SMF inline Bit8u BX_CPU_C_PREFIX get_DH(void) { return(DH); };

BX_SMF inline Bit16u BX_CPU_C_PREFIX get_AX(void) { return(AX); };
BX_SMF inline Bit16u BX_CPU_C_PREFIX get_BX(void) { return(BX); };
BX_SMF inline Bit16u BX_CPU_C_PREFIX get_CX(void) { return(CX); };
BX_SMF inline Bit16u BX_CPU_C_PREFIX get_DX(void) { return(DX); };


#if BX_CPU_LEVEL >= 2
  inline Boolean BX_CPU_C::real_mode(void) { return( !BX_CPU_THIS_PTR cr0.pe ); };
#endif

#if BX_CPU_LEVEL == 2
  inline Boolean BX_CPU_C::protected_mode(void) { return( BX_CPU_THIS_PTR cr0.pe ); };
#endif


#if BX_CPU_LEVEL >= 3
#  if BX_SUPPORT_V8086_MODE
  inline Boolean
BX_CPU_C::v8086_mode(void) {
  return(BX_CPU_THIS_PTR eflags.vm);
  }

  inline Boolean
BX_CPU_C::protected_mode(void) {
  return(BX_CPU_THIS_PTR cr0.pe && !BX_CPU_THIS_PTR eflags.vm);
  }
#  else
  inline Boolean
BX_CPU_C::v8086_mode(void) {
  return(0);
  }

  inline Boolean
BX_CPU_C::protected_mode(void) {
  return(BX_CPU_THIS_PTR cr0.pe);
  }
#  endif
#endif

    inline void
BX_CPU_C::set_CF(Boolean val) {
    BX_CPU_THIS_PTR lf_flags_status &= 0xfffff0;
    BX_CPU_THIS_PTR eflags.cf = val;
    }

    inline void
BX_CPU_C::set_AF(Boolean val) {
    BX_CPU_THIS_PTR lf_flags_status &= 0xfff0ff;
    BX_CPU_THIS_PTR eflags.af = val;
    }

    inline void
BX_CPU_C::set_ZF(Boolean val) {
    BX_CPU_THIS_PTR lf_flags_status &= 0xff0fff;
    BX_CPU_THIS_PTR eflags.zf = val;
    }

    inline void
BX_CPU_C::set_SF(Boolean val) {
    BX_CPU_THIS_PTR lf_flags_status &= 0xf0ffff;
    BX_CPU_THIS_PTR eflags.sf = val;
    }


    inline void
BX_CPU_C::set_OF(Boolean val) {
    BX_CPU_THIS_PTR lf_flags_status &= 0x0fffff;
    BX_CPU_THIS_PTR eflags.of = val;
    }

    inline void
BX_CPU_C::set_PF(Boolean val) {
    BX_CPU_THIS_PTR lf_flags_status &= 0xffff0f;
    BX_CPU_THIS_PTR lf_pf = val;
    }

    inline void
BX_CPU_C::set_PF_base(Bit8u val) {
    BX_CPU_THIS_PTR eflags.pf_byte = val;
    BX_CPU_THIS_PTR lf_flags_status = (BX_CPU_THIS_PTR lf_flags_status & 0xffff0f) | BX_LF_MASK_P;
    }


#define SET_FLAGS_OSZAPC_8(op1, op2, result, ins) { \
    BX_CPU_THIS_PTR oszapc.op1_8 = op1; \
    BX_CPU_THIS_PTR oszapc.op2_8 = op2; \
    BX_CPU_THIS_PTR oszapc.result_8 = result; \
    BX_CPU_THIS_PTR oszapc.instr = ins; \
    BX_CPU_THIS_PTR lf_flags_status = BX_LF_MASK_OSZAPC; \
    }

#define SET_FLAGS_OSZAPC_8_CF(op1, op2, result, ins, last_CF) { \
    BX_CPU_THIS_PTR oszapc.op1_8 = op1; \
    BX_CPU_THIS_PTR oszapc.op2_8 = op2; \
    BX_CPU_THIS_PTR oszapc.result_8 = result; \
    BX_CPU_THIS_PTR oszapc.instr = ins; \
    BX_CPU_THIS_PTR oszapc.prev_CF = last_CF; \
    BX_CPU_THIS_PTR lf_flags_status = BX_LF_MASK_OSZAPC; \
    }

#define SET_FLAGS_OSZAPC_16(op1, op2, result, ins) { \
    BX_CPU_THIS_PTR oszapc.op1_16 = op1; \
    BX_CPU_THIS_PTR oszapc.op2_16 = op2; \
    BX_CPU_THIS_PTR oszapc.result_16 = result; \
    BX_CPU_THIS_PTR oszapc.instr = ins; \
    BX_CPU_THIS_PTR lf_flags_status = BX_LF_MASK_OSZAPC; \
    }

#define SET_FLAGS_OSZAPC_16_CF(op1, op2, result, ins, last_CF) { \
    BX_CPU_THIS_PTR oszapc.op1_16 = op1; \
    BX_CPU_THIS_PTR oszapc.op2_16 = op2; \
    BX_CPU_THIS_PTR oszapc.result_16 = result; \
    BX_CPU_THIS_PTR oszapc.instr = ins; \
    BX_CPU_THIS_PTR oszapc.prev_CF = last_CF; \
    BX_CPU_THIS_PTR lf_flags_status = BX_LF_MASK_OSZAPC; \
    }

#define SET_FLAGS_OSZAPC_32(op1, op2, result, ins) { \
    BX_CPU_THIS_PTR oszapc.op1_32 = op1; \
    BX_CPU_THIS_PTR oszapc.op2_32 = op2; \
    BX_CPU_THIS_PTR oszapc.result_32 = result; \
    BX_CPU_THIS_PTR oszapc.instr = ins; \
    BX_CPU_THIS_PTR lf_flags_status = BX_LF_MASK_OSZAPC; \
    }

#define SET_FLAGS_OSZAPC_32_CF(op1, op2, result, ins, last_CF) { \
    BX_CPU_THIS_PTR oszapc.op1_32 = op1; \
    BX_CPU_THIS_PTR oszapc.op2_32 = op2; \
    BX_CPU_THIS_PTR oszapc.result_32 = result; \
    BX_CPU_THIS_PTR oszapc.instr = ins; \
    BX_CPU_THIS_PTR oszapc.prev_CF = last_CF; \
    BX_CPU_THIS_PTR lf_flags_status = BX_LF_MASK_OSZAPC; \
    }


#define SET_FLAGS_OSZAP_8(op1, op2, result, ins) { \
    BX_CPU_THIS_PTR oszap.op1_8 = op1; \
    BX_CPU_THIS_PTR oszap.op2_8 = op2; \
    BX_CPU_THIS_PTR oszap.result_8 = result; \
    BX_CPU_THIS_PTR oszap.instr = ins; \
    BX_CPU_THIS_PTR lf_flags_status = (BX_CPU_THIS_PTR lf_flags_status & 0x00000f) | BX_LF_MASK_OSZAP; \
    }

#define SET_FLAGS_OSZAP_16(op1, op2, result, ins) { \
    BX_CPU_THIS_PTR oszap.op1_16 = op1; \
    BX_CPU_THIS_PTR oszap.op2_16 = op2; \
    BX_CPU_THIS_PTR oszap.result_16 = result; \
    BX_CPU_THIS_PTR oszap.instr = ins; \
    BX_CPU_THIS_PTR lf_flags_status = (BX_CPU_THIS_PTR lf_flags_status & 0x00000f) | BX_LF_MASK_OSZAP; \
    }

#define SET_FLAGS_OSZAP_32(op1, op2, result, ins) { \
    BX_CPU_THIS_PTR oszap.op1_32 = op1; \
    BX_CPU_THIS_PTR oszap.op2_32 = op2; \
    BX_CPU_THIS_PTR oszap.result_32 = result; \
    BX_CPU_THIS_PTR oszap.instr = ins; \
    BX_CPU_THIS_PTR lf_flags_status = (BX_CPU_THIS_PTR lf_flags_status & 0x00000f) | BX_LF_MASK_OSZAP; \
    }

#define SET_FLAGS_OxxxxC(new_of, new_cf) { \
    BX_CPU_THIS_PTR eflags.of = (Boolean) new_of; \
    BX_CPU_THIS_PTR eflags.cf = (Boolean) new_cf; \
    BX_CPU_THIS_PTR lf_flags_status &= 0x0ffff0; \
    /* ??? could also mark other bits undefined here */ \
    }



#if 0
    inline void
  SET_FLAGS_OSZAPC_8(Bit8u op1, Bit8u op2, Bit8u result, unsigned instr) {
    oszapc.op1_8 = op1;
    oszapc.op2_8 = op2;
    oszapc.result_8 = result;
    oszapc.instr = instr;
    lf_flags_status = BX_LF_MASK_OSZAPC;
    }

    inline void
  SET_FLAGS_OSZAPC_8_CF(Bit8u op1, Bit8u op2, Bit8u result, unsigned instr,
                         Boolean prev_CF) {
    oszapc.op1_8 = op1;
    oszapc.op2_8 = op2;
    oszapc.result_8 = result;
    oszapc.instr = instr;
    oszapc.prev_CF = prev_CF;
    lf_flags_status = BX_LF_MASK_OSZAPC;
    }

    inline void
  SET_FLAGS_OSZAPC_16(Bit16u op1, Bit16u op2, Bit16u result, unsigned instr) {
    oszapc.op1_16 = op1;
    oszapc.op2_16 = op2;
    oszapc.result_16 = result;
    oszapc.instr = instr;
    lf_flags_status = BX_LF_MASK_OSZAPC;
    }

    inline void
  SET_FLAGS_OSZAPC_16_CF(Bit16u op1, Bit16u op2, Bit16u result, unsigned instr,
                          Boolean prev_CF) {
    oszapc.op1_16 = op1;
    oszapc.op2_16 = op2;
    oszapc.result_16 = result;
    oszapc.instr = instr;
    oszapc.prev_CF = prev_CF;
    lf_flags_status = BX_LF_MASK_OSZAPC;
    }

    inline void
  SET_FLAGS_OSZAPC_32(Bit32u op1, Bit32u op2, Bit32u result, unsigned instr) {
    oszapc.op1_32 = op1;
    oszapc.op2_32 = op2;
    oszapc.result_32 = result;
    oszapc.instr = instr;
    lf_flags_status = BX_LF_MASK_OSZAPC;
    }

    inline void
  SET_FLAGS_OSZAPC_32_CF(Bit32u op1, Bit32u op2, Bit32u result, unsigned instr,
                          Boolean prev_CF) {
    oszapc.op1_32 = op1;
    oszapc.op2_32 = op2;
    oszapc.result_32 = result;
    oszapc.instr = instr;
    oszapc.prev_CF = prev_CF;
    lf_flags_status = BX_LF_MASK_OSZAPC;
    }


    inline void
  SET_FLAGS_OSZAP_8(Bit8u op1, Bit8u op2, Bit8u result, unsigned instr) {
    oszap.op1_8 = op1;
    oszap.op2_8 = op2;
    oszap.result_8 = result;
    oszap.instr = instr;
    lf_flags_status = (lf_flags_status & 0x00000f) | BX_LF_MASK_OSZAP;
    }

    inline void
  SET_FLAGS_OSZAP_16(Bit16u op1, Bit16u op2, Bit16u result, unsigned instr) {
    oszap.op1_16 = op1;
    oszap.op2_16 = op2;
    oszap.result_16 = result;
    oszap.instr = instr;
    lf_flags_status = (lf_flags_status & 0x00000f) | BX_LF_MASK_OSZAP;
    }

    inline void
  SET_FLAGS_OSZAP_32(Bit32u op1, Bit32u op2, Bit32u result, unsigned instr) {
    oszap.op1_32 = op1;
    oszap.op2_32 = op2;
    oszap.result_32 = result;
    oszap.instr = instr;
    lf_flags_status = (lf_flags_status & 0x00000f) | BX_LF_MASK_OSZAP;
    }

    inline void
  SET_FLAGS_OxxxxC(Boolean new_of, Boolean new_cf) {
    eflags.of = new_of;
    eflags.cf = new_cf;
    lf_flags_status &= 0x0ffff0;
    /* ??? could also mark other bits undefined here */
    }
#endif



extern const Boolean bx_parity_lookup[256];

#define BX_REPE_PREFIX  10
#define BX_REPNE_PREFIX 11



#define BX_TASK_FROM_JUMP         10
#define BX_TASK_FROM_CALL_OR_INT  11
#define BX_TASK_FROM_IRET         12


#if USE_INLINE
#  error "rehash this in cpu.h"
#  include "fetch.cc"
#endif

#endif  // #ifndef BX_CPU_H
