// Copyright Notice
// ================
// BOCHS is Copyright 1994-1999 by Kevin P. Lawton.
//
// BOCHS is commercial software.
//
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
//
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02452-3028
//
//     EMail:    bochs@world.std.com



#include "bochs.h"





  INLINE void
BX_CPU_C::JCXZ_Jb(BxInstruction_t *i)
{
  Bit8s offset;
  Bit32u temp_ECX;

  offset = (Bit8s) i->Ib;

  if (i->as_32)
    temp_ECX = ECX;
  else
    temp_ECX = CX;

  if ( temp_ECX == 0 ) {
    Bit32u new_EIP;

    new_EIP = EIP + offset;
    if (i->os_32==0)
      new_EIP &= 0x0000ffff;
#if BX_CPU_LEVEL >= 2
    if (protected_mode()) {
      if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("jcxz_jb: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        }
      }
#endif
    EIP = new_EIP;
    BX_INSTR_CNEAR_BRANCH_TAKEN(new_EIP);
    revalidate_prefetch_q();
    }
#if BX_INSTRUMENTATION
  else {
    BX_INSTR_CNEAR_BRANCH_NOT_TAKEN();
    }
#endif
}



  INLINE void
BX_CPU_C::LOOPNE_Jb(BxInstruction_t *i)
{
  Bit8s imm8;
  Bit32u count, new_EIP;

  imm8 = i->Ib;

#if BX_CPU_LEVEL >= 3
  if (i->as_32)
    count = ECX;
  else
#endif /* BX_CPU_LEVEL >= 3 */
    count = CX;

  count--;
  if ( (count!=0) && (get_ZF()==0) ) {

    new_EIP = EIP + imm8;
    if (i->os_32==0)
      new_EIP &= 0x0000ffff;
    if (protected_mode()) {
      if (new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled) {
        bx_panic("loopne_jb: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        }
      }
    BX_CPU_THIS_PTR eip = new_EIP;
    BX_INSTR_CNEAR_BRANCH_TAKEN(new_EIP);
    revalidate_prefetch_q();
    }
#if BX_INSTRUMENTATION
  else {
    BX_INSTR_CNEAR_BRANCH_NOT_TAKEN();
    }
#endif

  if (i->as_32)
    ECX--;
  else
    CX--;
}

  INLINE void
BX_CPU_C::LOOPE_Jb(BxInstruction_t *i)
{
  Bit8s imm8;
  Bit32u count, new_EIP;

  imm8 = i->Ib;

#if BX_CPU_LEVEL >= 3
  if (i->as_32)
    count = ECX;
  else
#endif /* BX_CPU_LEVEL >= 3 */
    count = CX;

  count--;
  if ( (count!=0) && get_ZF()) {

    new_EIP = EIP + imm8;
    if (i->os_32==0)
      new_EIP &= 0x0000ffff;
    if (protected_mode()) {
      if (new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled) {
        bx_panic("loope_jb: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        }
      }
    BX_CPU_THIS_PTR eip = new_EIP;
    BX_INSTR_CNEAR_BRANCH_TAKEN(new_EIP);
    revalidate_prefetch_q();
    }
#if BX_INSTRUMENTATION
  else {
    BX_INSTR_CNEAR_BRANCH_NOT_TAKEN();
    }
#endif

  if (i->as_32)
    ECX--;
  else
    CX--;
}

  INLINE void
BX_CPU_C::LOOP_Jb(BxInstruction_t *i)
{
  Bit8s imm8;
  Bit32u count, new_EIP;

  imm8 = i->Ib;

#if BX_CPU_LEVEL >= 3
  if (i->as_32)
    count = ECX;
  else
#endif /* BX_CPU_LEVEL >= 3 */
    count = CX;

  count--;
  if (count != 0) {

    new_EIP = EIP + imm8;
    if (i->os_32==0)
      new_EIP &= 0x0000ffff;
    if (protected_mode()) {
      if (new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled) {
        bx_panic("loop_jb: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        }
      }
    BX_CPU_THIS_PTR eip = new_EIP;
    BX_INSTR_CNEAR_BRANCH_TAKEN(new_EIP);
    revalidate_prefetch_q();
    }
#if BX_INSTRUMENTATION
  else {
    BX_INSTR_CNEAR_BRANCH_NOT_TAKEN();
    }
#endif

  if (i->as_32)
    ECX--;
  else
    CX--;
}
