// Copyright Notice
// ================
// BOCHS is Copyright 1994-1999 by Kevin P. Lawton.
//
// BOCHS is commercial software.
//
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
//
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02452-3028
//
//     EMail:    bochs@world.std.com



#include "bochs.h"




#define INDEX_PULSE_CYCLE 10

#if BX_USE_HD_SMF
bx_hard_drive_c bx_hard_drive;
#define this (&bx_hard_drive)
#endif


static unsigned char model_no[41] =
  "Generic 1234                            ";
static unsigned max_multiple_sectors  = 0; // was 0x3f
static unsigned curr_multiple_sectors = 0; // was 0x3f

//static unsigned im_here = 0;

bx_hard_drive_c::bx_hard_drive_c(void)
{
  // nothing for now
}

bx_hard_drive_c::~bx_hard_drive_c(void)
{
  // nothing for now
}



  void
bx_hard_drive_c::init(bx_devices_c *d, bx_cmos_c *cmos)
{
  struct stat stat_buf;
  int ret;

  BX_HD_THIS devices = d;

  /* HARD DRIVE 0 */

  BX_HD_THIS devices->register_irq(14, "Hard Drive 0");
  for (unsigned addr=0x01F0; addr<=0x01F7; addr++) {
    BX_HD_THIS devices->register_io_read_handler(this, read_handler,
                                        addr, "Hard Drive 0");
    BX_HD_THIS devices->register_io_write_handler(this, write_handler,
                                        addr, "Hard Drive 0");
    }

  /* open hard drive image file */
  BX_HD_THIS s.hard_drive.fd = open(bx_options.diskc.path, O_RDWR
#ifdef O_BINARY
                                    | O_BINARY
#endif
                                   );

  if (BX_HD_THIS s.hard_drive.fd < 0) {
    bx_panic("could not open hard drive image file '%s'!\n",
      bx_options.diskc.path);
    }

  /* look at size of image file to calculate disk geometry */
  ret = fstat(BX_HD_THIS s.hard_drive.fd, &stat_buf);
  if (ret) {
    perror("fstat'ing hard drive image file");
    bx_panic("fstat() returns error!\n");
    }

  // generate CMOS values for hard drive if not using a CMOS image
  if (!bx_options.cmos.cmosImage) {
    cmos->s.reg[0x12] = 0x00; // start out with: no drive 0, no drive 1

    // Flag drive type as Fh, use extended CMOS location as real type
    cmos->s.reg[0x12] = (cmos->s.reg[0x12] & 0x0f) | 0xf0;
    cmos->s.reg[0x19] = 47; // user definable type
    cmos->s.reg[0x1b] = (bx_options.diskc.cylinders & 0x00ff);
    cmos->s.reg[0x1c] = (bx_options.diskc.cylinders & 0xff00) >> 8;
    cmos->s.reg[0x1d] = (bx_options.diskc.heads);
    cmos->s.reg[0x1e] = 0xff; // -1
    cmos->s.reg[0x1f] = 0xff; // -1
    cmos->s.reg[0x20] = 0xc0 | ((bx_options.diskc.heads > 8) << 3);
    cmos->s.reg[0x21] = cmos->s.reg[0x1b];
    cmos->s.reg[0x22] = cmos->s.reg[0x1c];
    cmos->s.reg[0x23] = bx_options.diskc.spt;

    if ( bx_options.bootdrive[0] == 'c' ) {
      // system boot sequence C:, A:
      cmos->s.reg[0x2d] &= 0xdf;
      }
    else { // 'a'
      // system boot sequence A:, C:
      cmos->s.reg[0x2d] |= 0x20;
      }
    }

  BX_HD_THIS s.hard_drive.cylinders = bx_options.diskc.cylinders;
  BX_HD_THIS s.hard_drive.heads     = bx_options.diskc.heads;
  BX_HD_THIS s.hard_drive.sectors   = bx_options.diskc.spt;

  //switch (stat_buf.st_size) {
  //  }

  BX_HD_THIS s.controller.status.busy           = 0;
  BX_HD_THIS s.controller.status.drive_ready    = 1;
  BX_HD_THIS s.controller.status.write_fault    = 0;
  BX_HD_THIS s.controller.status.seek_complete  = 1;
  BX_HD_THIS s.controller.status.drq            = 0;
  BX_HD_THIS s.controller.status.corrected_data = 0;
  BX_HD_THIS s.controller.status.index_pulse    = 0;
  BX_HD_THIS s.controller.status.index_pulse_count = 0;
  BX_HD_THIS s.controller.status.err            = 0;

  BX_HD_THIS s.controller.error_register = 0x01; // diagnostic code: no error
  BX_HD_THIS s.controller.drive_select   = 0;
  BX_HD_THIS s.controller.head_no        = 0;
  BX_HD_THIS s.controller.sector_count   = 1;
  BX_HD_THIS s.controller.sector_no      = 1;
  BX_HD_THIS s.controller.cylinder_no    = 0;
  BX_HD_THIS s.controller.current_command = 0x00;
  BX_HD_THIS s.controller.buffer_index = 0;

  BX_HD_THIS s.controller.control.reset       = 0;
  BX_HD_THIS s.controller.control.disable_irq = 0;
}



  // static IO port read callback handler
  // redirects to non-static class handler to avoid virtual functions

  Bit32u
bx_hard_drive_c::read_handler(void *this_ptr, Bit32u address, unsigned io_len)
{
#if !BX_USE_HD_SMF
  bx_hard_drive_c *class_ptr = (bx_hard_drive_c *) this_ptr;

  return( class_ptr->read(address, io_len) );
}


  Bit32u
bx_hard_drive_c::read(Bit32u address, unsigned io_len)
{
#else
  UNUSED(this_ptr);
#endif  // !BX_USE_HD_SMF
  Bit8u value8;
  Bit16u value16;

  if (io_len==2 && address!=0x1f0) {
    bx_panic("disk: non-byte IO read to %04x\n", (unsigned) address);
    }

  if (bx_dbg.disk)
    bx_printf("disk: IO read from %04x\n", (unsigned) address);

  switch (address) {
    case 0x1f0: // hard disk data (16bit)
      switch (BX_HD_THIS s.controller.current_command) {
        case 0x20: // read sectors
          if (io_len != 2) {
            bx_panic("disk: non-word IO read from %04x\n",
                     (unsigned) address);
            }
          if (BX_HD_THIS s.controller.buffer_index >= 512)
            bx_panic("disk: IO read(1f0): buffer_index >= 512\n");
          value16  = BX_HD_THIS s.controller.buffer[BX_HD_THIS s.controller.buffer_index];
          value16 |= (BX_HD_THIS s.controller.buffer[BX_HD_THIS s.controller.buffer_index+1] << 8);
          BX_HD_THIS s.controller.buffer_index += 2;

          // if buffer completely read
          if (BX_HD_THIS s.controller.buffer_index >= 512) {
            // update sector count, sector number, cylinder,
            // drive, head, status
            // if there are more sectors, read next one in...
            //
            BX_HD_THIS s.controller.buffer_index = 0;
            /* sector count of 0 means 256, it'll wrap */
            BX_HD_THIS s.controller.sector_count--;
            BX_HD_THIS s.controller.sector_no++;
            if (BX_HD_THIS s.controller.sector_no > BX_HD_THIS s.hard_drive.sectors) {
              BX_HD_THIS s.controller.sector_no = 1;
              BX_HD_THIS s.controller.head_no++;
              if (BX_HD_THIS s.controller.head_no >= BX_HD_THIS s.hard_drive.heads) {
                BX_HD_THIS s.controller.head_no = 0;
                BX_HD_THIS s.controller.cylinder_no++;
                if (BX_HD_THIS s.controller.cylinder_no >= BX_HD_THIS s.hard_drive.cylinders)
                  BX_HD_THIS s.controller.cylinder_no = BX_HD_THIS s.hard_drive.cylinders - 1;
                }
              }

            BX_HD_THIS s.controller.status.busy = 0;
            BX_HD_THIS s.controller.status.drive_ready = 1;
            BX_HD_THIS s.controller.status.write_fault = 0;
            if (bx_options.newHardDriveSupport)
              BX_HD_THIS s.controller.status.seek_complete = 1;
            else
              BX_HD_THIS s.controller.status.seek_complete = 0;
            BX_HD_THIS s.controller.status.corrected_data = 0;
            BX_HD_THIS s.controller.status.err = 0;

            if (BX_HD_THIS s.controller.sector_count==0) {
              BX_HD_THIS s.controller.status.drq = 0;
              }
            else { /* read next one into controller buffer */
              unsigned long logical_sector;
              int ret;

              BX_HD_THIS s.controller.status.drq = 1;
              BX_HD_THIS s.controller.status.seek_complete = 1;

              logical_sector = (BX_HD_THIS s.controller.cylinder_no * BX_HD_THIS s.hard_drive.heads *
                                BX_HD_THIS s.hard_drive.sectors) +
                               (BX_HD_THIS s.controller.head_no * BX_HD_THIS s.hard_drive.sectors) +
                               (BX_HD_THIS s.controller.sector_no - 1);

              if (logical_sector >= (BX_HD_THIS s.hard_drive.cylinders *
                  BX_HD_THIS s.hard_drive.heads * BX_HD_THIS s.hard_drive.sectors))
                bx_panic("disk: read sectors: out of bounds\n");
              ret = lseek(BX_HD_THIS s.hard_drive.fd, logical_sector * 512, SEEK_SET);

              if (ret < 0)
                bx_panic("disk: could lseek() hard drive image file\n");
              ret = ::read(BX_HD_THIS s.hard_drive.fd,
                           (bx_ptr_t) BX_HD_THIS s.controller.buffer, 512);
              if (ret < 512) {
                bx_printf("logical sector was %u\n", (unsigned) logical_sector);
                bx_panic("disk: could not read() hard drive image file\n");
                }

              BX_HD_THIS s.controller.buffer_index = 0;
              if (!BX_HD_THIS s.controller.control.disable_irq)
                BX_HD_THIS devices->pic->trigger_irq(14);
              }
            }
          return(value16);
          break;

        case 0xec:    // Drive ID Command
          if (bx_options.newHardDriveSupport) {
            unsigned index;
            Bit32u value32;

            BX_HD_THIS s.controller.status.busy = 0;
            BX_HD_THIS s.controller.status.drive_ready = 1;
            BX_HD_THIS s.controller.status.write_fault = 0;
            BX_HD_THIS s.controller.status.seek_complete = 1;
            BX_HD_THIS s.controller.status.corrected_data = 0;
            BX_HD_THIS s.controller.status.err = 0;

            index = BX_HD_THIS s.controller.buffer_index;
            value32 = BX_HD_THIS s.controller.buffer[index];
            index++;
            if (io_len >= 2) {
              value32 |= (BX_HD_THIS s.controller.buffer[index] << 8);
              index++;
              }
            if (io_len == 4) {
              value32 |= (BX_HD_THIS s.controller.buffer[index] << 16);
              value32 |= (BX_HD_THIS s.controller.buffer[index+1] << 24);
              index += 2;
              }
            BX_HD_THIS s.controller.buffer_index = index;

            if (BX_HD_THIS s.controller.buffer_index >= 512) {
              BX_HD_THIS s.controller.status.drq = 0;
              bx_printf ("disk: Read all drive ID Bytes ...\n");
              }
            return (value32);
            }
          else
            bx_panic("disk: IO read(1f0h): current command is %02xh\n",
              (unsigned) BX_HD_THIS s.controller.current_command);

        default:
          bx_panic("disk: IO read(1f0h): current command is %02xh\n",
            (unsigned) BX_HD_THIS s.controller.current_command);
        }
      break;

    case 0x1f1: // hard disk error register
      BX_HD_THIS s.controller.status.err = 0;
      value8 = BX_HD_THIS s.controller.error_register;
      return(value8);
      break;

    case 0x1f2: // hard disk sector count
      if (BX_HD_THIS s.controller.current_command==0x20 ||
          BX_HD_THIS s.controller.current_command==0x30) {
        value8 = BX_HD_THIS s.controller.sector_count;
        return(value8);
        }
      bx_panic("disk: IO read(0x1f2): current command not read/write\n");
      break;

    case 0x1f3: // sector number
      value8 = BX_HD_THIS s.controller.sector_no;
      return(value8);
    case 0x1f4: // cylinder low
      value8 = (BX_HD_THIS s.controller.cylinder_no & 0x00ff);
      return(value8);
    case 0x1f5: // cylinder high
      value8 = BX_HD_THIS s.controller.cylinder_no >> 8;
      return(value8);

    case 0x1f6: // hard disk drive and head register
      value8 = (1 << 7) | // extended data field for ECC
               (0 << 7) | // 1=LBA mode, 0=CHSmode
               (1 << 5) | // 01b = 512 sector size
               (BX_HD_THIS s.controller.drive_select << 4) |
               (BX_HD_THIS s.controller.head_no << 0);
      return(value8);
      break;

    case 0x1f7: // Hard Disk Status
    case 0x3f6: // Hard Disk Alternate Status
      value8 = (
        (BX_HD_THIS s.controller.status.busy << 7) |
        (BX_HD_THIS s.controller.status.drive_ready << 6) |
        (BX_HD_THIS s.controller.status.write_fault << 5) |
        (BX_HD_THIS s.controller.status.seek_complete << 4) |
        (BX_HD_THIS s.controller.status.drq << 3) |
        (BX_HD_THIS s.controller.status.corrected_data << 2) |
        (BX_HD_THIS s.controller.status.index_pulse << 1) |
        (BX_HD_THIS s.controller.status.err) );
      BX_HD_THIS s.controller.status.index_pulse_count++;
      BX_HD_THIS s.controller.status.index_pulse = 0;
      if (BX_HD_THIS s.controller.status.index_pulse_count >= INDEX_PULSE_CYCLE) {
        BX_HD_THIS s.controller.status.index_pulse = 1;
        BX_HD_THIS s.controller.status.index_pulse_count = 0;
        }
      return(value8);
      break;

    case 0x3f7: // Hard Disk Address Register
      // Obsolete and unsupported register.  Not driven by hard
      // disk controller.  Report all 1's.  If floppy controller
      // is handling this address, it will call this function
      // set/clear D7 (the only bit it handles), then return
      // the combined value
      return(0xff);
      break;

    default:
      bx_panic("hard drive: io read to address %x unsupported\n",
        (unsigned) address);
    }

  bx_panic("hard drive: shouldnt get here!\n");
  return(0);
}


  // static IO port write callback handler
  // redirects to non-static class handler to avoid virtual functions

  void
bx_hard_drive_c::write_handler(void *this_ptr, Bit32u address, Bit32u value, unsigned io_len)
{
#if !BX_USE_HD_SMF
  bx_hard_drive_c *class_ptr = (bx_hard_drive_c *) this_ptr;

  class_ptr->write(address, value, io_len);
}

  void
bx_hard_drive_c::write(Bit32u address, Bit32u value, unsigned io_len)
{
#else
  UNUSED(this_ptr);
#endif  // !BX_USE_HD_SMF
  unsigned long logical_sector;
  int ret;
  Boolean prev_control_reset;

  if (io_len==2 && address!=0x1f0) {
    bx_panic("disk: non-byte IO write to %04x\n", (unsigned) address);
    }

  if (bx_dbg.disk)
    bx_printf("disk: IO write to %04x = %02x\n",
      (unsigned) address, (unsigned) value);

//bx_printf("disk: IO write to %04x = %02x\n",
//      (unsigned) address, (unsigned) value);

  switch (address) {
    case 0x1f0:
      if (io_len != 2) {
        bx_panic("disk: non-word IO read from %04x\n", (unsigned) address);
        }
      switch (BX_HD_THIS s.controller.current_command) {
        case 0x30:
          if (BX_HD_THIS s.controller.buffer_index >= 512)
            bx_panic("disk: IO write(1f0): buffer_index >= 512\n");
          BX_HD_THIS s.controller.buffer[BX_HD_THIS s.controller.buffer_index] = value;
          BX_HD_THIS s.controller.buffer[BX_HD_THIS s.controller.buffer_index+1] = (value >> 8);
          BX_HD_THIS s.controller.buffer_index += 2;

          /* if buffer completely writtten */
          if (BX_HD_THIS s.controller.buffer_index >= 512) {
            unsigned long logical_sector;
            int ret;

            logical_sector = (BX_HD_THIS s.controller.cylinder_no * BX_HD_THIS s.hard_drive.heads *
                              BX_HD_THIS s.hard_drive.sectors) +
                             (BX_HD_THIS s.controller.head_no * BX_HD_THIS s.hard_drive.sectors) +
                             (BX_HD_THIS s.controller.sector_no - 1);

            if (logical_sector >= (BX_HD_THIS s.hard_drive.cylinders *
                BX_HD_THIS s.hard_drive.heads * BX_HD_THIS s.hard_drive.sectors))
              bx_panic("disk: write sectors: out of bounds\n");
            ret = lseek(BX_HD_THIS s.hard_drive.fd, logical_sector * 512, SEEK_SET);
            if (ret < 0)
              bx_panic("disk: could lseek() hard drive image file\n");

            ret = ::write(BX_HD_THIS s.hard_drive.fd,
                          (bx_ptr_t) BX_HD_THIS s.controller.buffer, 512);
            if (ret < 512)
              bx_panic("disk: could not write() hard drive image file\n");

            BX_HD_THIS s.controller.buffer_index = 0;

            /* update sector count, sector number, cylinder,
             * drive, head, status
             * if there are more sectors, read next one in...
             */
            /* sector count of 0 means 256, it'll wrap */
            BX_HD_THIS s.controller.sector_count--;
            BX_HD_THIS s.controller.sector_no++;
            if (BX_HD_THIS s.controller.sector_no > BX_HD_THIS s.hard_drive.sectors) {
              BX_HD_THIS s.controller.sector_no = 1;
              BX_HD_THIS s.controller.head_no++;
              if (BX_HD_THIS s.controller.head_no >= BX_HD_THIS s.hard_drive.heads) {
                BX_HD_THIS s.controller.head_no = 0;
                BX_HD_THIS s.controller.cylinder_no++;
                if (BX_HD_THIS s.controller.cylinder_no >= BX_HD_THIS s.hard_drive.cylinders)
                  BX_HD_THIS s.controller.cylinder_no = BX_HD_THIS s.hard_drive.cylinders - 1;
                }
              }

            /* When the write is complete, controller clears the DRQ bit and
             * sets the BSY bit.
             * If at least one more sector is to be written, controller sets DRQ bit,
             * clears BSY bit, and issues IRQ 14
             */

            if (BX_HD_THIS s.controller.sector_count!=0) {
              BX_HD_THIS s.controller.status.busy = 0;
              BX_HD_THIS s.controller.status.drive_ready = 1;
              BX_HD_THIS s.controller.status.drq = 1;
              BX_HD_THIS s.controller.status.corrected_data = 0;
              BX_HD_THIS s.controller.status.err = 0;
              }
            else { /* no more sectors to write */
              BX_HD_THIS s.controller.status.busy = 0;
              BX_HD_THIS s.controller.status.drive_ready = 1;
              BX_HD_THIS s.controller.status.drq = 0;
              BX_HD_THIS s.controller.status.err = 0;
              BX_HD_THIS s.controller.status.corrected_data = 0;
              }
            if (!BX_HD_THIS s.controller.control.disable_irq)
              BX_HD_THIS devices->pic->trigger_irq(14);
            }
          break;

        default:
          bx_panic("disk: IO write(1f0h): current command is %02xh\n",
            (unsigned) BX_HD_THIS s.controller.current_command);
        }
      break;

    case 0x1f1: /* hard disk write precompensation */
      if (bx_dbg.disk) {
        if (value == 0xff)
          bx_printf("disk: no precompensation\n");
        else
          bx_printf("disk: precompensation value %02x\n", (unsigned) value);
        }
      break;

    case 0x1f2: /* hard disk sector count */
      BX_HD_THIS s.controller.sector_count = value;
      if (bx_dbg.disk)
        bx_printf("disk: sector count = %u\n", (unsigned) value);
      break;

    case 0x1f3: /* hard disk sector number */
      BX_HD_THIS s.controller.sector_no = value;
      if (bx_dbg.disk)
        bx_printf("disk: sector number = %u\n", (unsigned) value);
      break;

    case 0x1f4: /* hard disk cylinder low */
      BX_HD_THIS s.controller.cylinder_no = (BX_HD_THIS s.controller.cylinder_no & 0xff00) | value;
      if (bx_dbg.disk)
        bx_printf("disk: cylinder low = %02xh\n", (unsigned) value);
      break;

    case 0x1f5: /* hard disk cylinder high */
      BX_HD_THIS s.controller.cylinder_no = (BX_HD_THIS s.controller.cylinder_no & 0x00ff) |
                               (value << 8);
      if (bx_dbg.disk)
        bx_printf("disk: cylinder high = %02xh\n", (unsigned) value);
      break;

    case 0x1f6: // hard disk drive and head register
      // b7 1
      // b6 1=LBA mode, 0=CHS mode (LBA not supported)
      // b5 1
      // b4: DRV
      // b3..0 HD3..HD0
      if ( (value & 0xe0) != 0xa0 ) // 101xxxxx
        bx_printf("disk: IO write 1f6 (%02x): not 101xxxxxb\n", (unsigned) value);
      BX_HD_THIS s.controller.drive_select = (value >> 4) & 0x01;
      BX_HD_THIS s.controller.head_no = (value & 0x0f);
      break;

    case 0x1f7: // hard disk command
      if (BX_HD_THIS s.controller.status.busy)
        bx_panic("hard disk: command sent, controller BUSY\n");
      if ( (value & 0xf0) == 0x10 )
        value = 0x10;
      switch (value) {

        case 0x10: // calibrate drive
          if (BX_HD_THIS s.controller.drive_select != 0) {
            BX_HD_THIS s.controller.error_register = 0x02; // Track 0 not found
            BX_HD_THIS s.controller.status.busy = 0;
            BX_HD_THIS s.controller.status.drive_ready = 1;
            BX_HD_THIS s.controller.status.seek_complete = 0;
            BX_HD_THIS s.controller.status.drq = 0;
            BX_HD_THIS s.controller.status.err = 1;
            if (!BX_HD_THIS s.controller.control.disable_irq)
              BX_HD_THIS devices->pic->trigger_irq(14);
            bx_printf("disk: calibrate drive != 0\n");
            break;
            }

          /* move head to cylinder 0, issue IRQ 14 */
          BX_HD_THIS s.controller.error_register = 0;
          BX_HD_THIS s.controller.cylinder_no = 0;
          BX_HD_THIS s.controller.status.busy = 0;
          BX_HD_THIS s.controller.status.drive_ready = 1;
          BX_HD_THIS s.controller.status.seek_complete = 1;
          BX_HD_THIS s.controller.status.drq = 0;
          BX_HD_THIS s.controller.status.err = 0;
          if (!BX_HD_THIS s.controller.control.disable_irq)
            BX_HD_THIS devices->pic->trigger_irq(14);
          break;

        case 0x20: /* read multiple sectors, with retries */
          /* update sector_no, always points to current sector
           * after each sector is read to buffer, DRQ bit set and issue IRQ 14
           * if interrupt handler transfers all data words into main memory,
           * and more sectors to read, then set BSY bit again, clear DRQ and
           * read next sector into buffer
           * sector count of 0 means 256 sectors
           */

          BX_HD_THIS s.controller.current_command = value;

          logical_sector = (BX_HD_THIS s.controller.cylinder_no * BX_HD_THIS s.hard_drive.heads *
                            BX_HD_THIS s.hard_drive.sectors) +
                           (BX_HD_THIS s.controller.head_no * BX_HD_THIS s.hard_drive.sectors) +
                           (BX_HD_THIS s.controller.sector_no - 1);

          if (logical_sector >=
              (BX_HD_THIS s.hard_drive.cylinders * BX_HD_THIS s.hard_drive.heads * BX_HD_THIS s.hard_drive.sectors)) {
            bx_panic("disk: read sectors: out of bounds\n");
            }
          ret = lseek(BX_HD_THIS s.hard_drive.fd, logical_sector * 512, SEEK_SET);

          if (ret < 0) {
            bx_panic("disk: could not lseek() hard drive image file\n");
            }
          ret = ::read(BX_HD_THIS s.hard_drive.fd,
                       (bx_ptr_t) BX_HD_THIS s.controller.buffer, 512);
          if (ret < 512) {
            bx_printf("logical sector was %u\n", (unsigned) logical_sector);
            bx_panic("disk: could not read() hard drive image file\n");
            }

          BX_HD_THIS s.controller.error_register = 0;
          BX_HD_THIS s.controller.status.busy  = 0;
          BX_HD_THIS s.controller.status.drive_ready = 1;
          BX_HD_THIS s.controller.status.seek_complete = 1;
          BX_HD_THIS s.controller.status.drq   = 1;
          BX_HD_THIS s.controller.status.corrected_data = 0;
          BX_HD_THIS s.controller.status.err   = 0;
          BX_HD_THIS s.controller.buffer_index = 0;
          if (!BX_HD_THIS s.controller.control.disable_irq)
            BX_HD_THIS devices->pic->trigger_irq(14);
          break;

        case 0x30: /* write multiple sectors, with retries */
          /* update sector_no, always points to current sector
           * after each sector is read to buffer, DRQ bit set and issue IRQ 14
           * if interrupt handler transfers all data words into main memory,
           * and more sectors to read, then set BSY bit again, clear DRQ and
           * read next sector into buffer
           * sector count of 0 means 256 sectors
           */

          if (BX_HD_THIS s.controller.status.busy) {
            bx_panic("disk: write command: BSY bit set\n");
            }
          BX_HD_THIS s.controller.current_command = value;

          // implicit seek done :^)
          BX_HD_THIS s.controller.error_register = 0;
          BX_HD_THIS s.controller.status.busy = 0;
          // BX_HD_THIS s.controller.status.drive_ready = 1;
          BX_HD_THIS s.controller.status.seek_complete = 1;
          BX_HD_THIS s.controller.status.drq = 1;
          BX_HD_THIS s.controller.status.err   = 0;
          BX_HD_THIS s.controller.buffer_index = 0;
          break;

        case 0x90: // Drive Diagnostic
          if (BX_HD_THIS s.controller.status.busy) {
            bx_panic("disk: diagnostic command: BSY bit set\n");
            }
          BX_HD_THIS s.controller.error_register = 0x81; // Drive 1 failed, no error on drive 0
          // BX_HD_THIS s.controller.status.busy = 0; // not needed
          BX_HD_THIS s.controller.status.drq = 0;
          BX_HD_THIS s.controller.status.err = 0;
          break;

        case 0x91: // initialize drive parameters
          if (BX_HD_THIS s.controller.status.busy) {
            bx_panic("disk: init drive parameters command: BSY bit set\n");
            }
          // sets logical geometry of specified drive
          bx_printf("initialize drive params\n");
          bx_printf("  sector count = %u\n",
            (unsigned) BX_HD_THIS s.controller.sector_count);
          bx_printf("  drive select = %u\n",
            (unsigned) BX_HD_THIS s.controller.drive_select);
          bx_printf("  head number = %u\n",
            (unsigned) BX_HD_THIS s.controller.head_no);
          if (BX_HD_THIS s.controller.drive_select != 0) {
            bx_panic("disk: init drive params: drive != 0\n");
            //BX_HD_THIS s.controller.error_register = 0x12;
            BX_HD_THIS s.controller.status.busy = 0;
            BX_HD_THIS s.controller.status.drive_ready = 1;
            BX_HD_THIS s.controller.status.drq = 0;
            BX_HD_THIS s.controller.status.err = 0;
            if (!BX_HD_THIS s.controller.control.disable_irq)
              BX_HD_THIS devices->pic->trigger_irq(14);
            break;
            }
          if (BX_HD_THIS s.controller.sector_count != BX_HD_THIS s.hard_drive.sectors)
            bx_panic("disk: init drive params: sector count doesnt match\n");
          if ( BX_HD_THIS s.controller.head_no != (BX_HD_THIS s.hard_drive.heads-1) )
            bx_panic("disk: init drive params: head number doesn't match\n");
          BX_HD_THIS s.controller.status.busy = 0;
          BX_HD_THIS s.controller.status.drive_ready = 1;
          BX_HD_THIS s.controller.status.drq = 0;
          BX_HD_THIS s.controller.status.err = 0;
          if (!BX_HD_THIS s.controller.control.disable_irq)
            BX_HD_THIS devices->pic->trigger_irq(14);
          break;

        case 0xec: // Get Drive Info
          if (bx_options.newHardDriveSupport) {
            bx_printf ("disk: Drive ID Command issued : 0xec \n");

            if (BX_HD_THIS s.controller.drive_select) {
              // 2nd drive not supported yet
              command_aborted(value);
              break;
              }
            BX_HD_THIS s.controller.current_command = value;
            BX_HD_THIS s.controller.error_register = 0;
            BX_HD_THIS s.controller.status.busy  = 0;
            BX_HD_THIS s.controller.status.drive_ready = 1;
            BX_HD_THIS s.controller.status.seek_complete = 1;
            BX_HD_THIS s.controller.status.drq   = 1;
            BX_HD_THIS s.controller.status.corrected_data = 0;
            BX_HD_THIS s.controller.status.err   = 0;
            BX_HD_THIS s.controller.buffer_index = 0;
            if (!BX_HD_THIS s.controller.control.disable_irq)
              BX_HD_THIS devices->pic->trigger_irq(14);
            identify_drive(BX_HD_THIS s.controller.drive_select);
            }
          else
            command_aborted(value);
          break;

        case 0x40: //
          if (bx_options.newHardDriveSupport) {
            bx_printf ("disk: Verify Command : 0x40 ! \n");
            BX_HD_THIS s.controller.status.busy = 0;
            BX_HD_THIS s.controller.status.drive_ready = 1;
            BX_HD_THIS s.controller.status.drq = 0;
            BX_HD_THIS s.controller.status.err = 0;
            if (!BX_HD_THIS s.controller.control.disable_irq)
              BX_HD_THIS devices->pic->trigger_irq(14);
            }
          else
            command_aborted(value);
          break;

        // ATAPI commands
        case 0x08: // ATAPI soft reset command
        case 0xa0: // send a packet command
        case 0xa1: // identify ATAPI device
        // non-standard commands
        case 0xf0: // Exabyte enable nest command
          command_aborted(value);
          break;


        default:
          bx_panic("IO write(1f7h): command %02x\n", (unsigned) value);
        }
      break;

    case 0x3f6: // hard disk adapter control
      prev_control_reset = BX_HD_THIS s.controller.control.reset;
      BX_HD_THIS s.controller.control.reset          = value & 0x04;
      BX_HD_THIS s.controller.control.disable_irq    = value & 0x02;
      //fprintf(stderr, "# hard drive: adpater control reg: reset controller = %d\n",
      //  (unsigned) (BX_HD_THIS s.controller.control.reset) ? 1 : 0);
      //fprintf(stderr, "# hard drive: adpater control reg: disable_irq(14) = %d\n",
      //  (unsigned) (BX_HD_THIS s.controller.control.disable_irq) ? 1 : 0);
      if (!prev_control_reset && BX_HD_THIS s.controller.control.reset) {
        // transition from 0 to 1 causes all drives to reset
        bx_printf("hard drive: RESET\n");
        BX_HD_THIS s.controller.status.busy           = 0;
        BX_HD_THIS s.controller.status.drive_ready    = 1;
        BX_HD_THIS s.controller.status.write_fault    = 0;
        BX_HD_THIS s.controller.status.seek_complete  = 1;
        BX_HD_THIS s.controller.status.drq            = 0;
        BX_HD_THIS s.controller.status.corrected_data = 0;
        BX_HD_THIS s.controller.status.err            = 0;

        BX_HD_THIS s.controller.error_register = 0x01; // diagnostic code: no error
        BX_HD_THIS s.controller.drive_select   = 0;
        BX_HD_THIS s.controller.head_no        = 0;
        BX_HD_THIS s.controller.sector_count   = 1;
        BX_HD_THIS s.controller.sector_no      = 1;
        BX_HD_THIS s.controller.cylinder_no    = 0;
        BX_HD_THIS s.controller.current_command = 0x00;
        BX_HD_THIS s.controller.buffer_index = 0;

        BX_HD_THIS s.controller.control.disable_irq = 0;
        }
      break;

    default:
      bx_panic("hard drive: io write to address %x = %02x\n",
        (unsigned) address, (unsigned) value);
    }
}

  void
bx_hard_drive_c::close_harddrive(void)
{
  if (BX_HD_THIS s.hard_drive.fd > -1) {
    close(BX_HD_THIS s.hard_drive.fd);
    }
}

  void
bx_hard_drive_c::identify_drive(unsigned drive)
{
  unsigned i;
  Bit32u temp32;
  Bit16u temp16;

if (drive) {
  bx_panic("disk: identify_drive: drive=1\n");
  }

  // Identify Drive command return values definition
  //
  // This code is rehashed from some that was donated.
  // I'm using ANSI X3.221-1994, AT Attachment Interface for Disk Drives
  // and X3T10 2008D Working Draft for ATA-3


  // Word 0: general config bit-significant info
  //   Note: bits 1-5 and 8-14 are now "Vendor specific (obsolete)"
  //   bit 15: 0=ATA device
  //           1=ATAPI device
  //   bit 14: 1=format speed tolerance gap required
  //   bit 13: 1=track offset option available
  //   bit 12: 1=data strobe offset option available
  //   bit 11: 1=rotational speed tolerance is > 0,5% (typo?)
  //   bit 10: 1=disk transfer rate > 10Mbs
  //   bit  9: 1=disk transfer rate > 5Mbs but <= 10Mbs
  //   bit  8: 1=disk transfer rate <= 5Mbs
  //   bit  7: 1=removable cartridge drive
  //   bit  6: 1=fixed drive
  //   bit  5: 1=spindle motor control option implemented
  //   bit  4: 1=head switch time > 15 usec
  //   bit  3: 1=not MFM encoded
  //   bit  2: 1=soft sectored
  //   bit  1: 1=hard sectored
  //   bit  0: 0=reserved
  BX_HD_THIS s.id_drive[0] = 0x0040;

  // Word 1: number of user-addressable cylinders in
  //   default translation mode.  If the value in words 60-61
  //   exceed 16,515,072, this word shall contain 16,383.
  BX_HD_THIS s.id_drive[1] = BX_HD_THIS s.hard_drive.cylinders;

  // Word 2: reserved
  BX_HD_THIS s.id_drive[2] = 0;

  // Word 3: number of user-addressable heads in default
  //   translation mode
  BX_HD_THIS s.id_drive[3] = BX_HD_THIS s.hard_drive.heads;

  // Word 4: # unformatted bytes per translated track in default xlate mode
  // Word 5: # unformatted bytes per sector in default xlated mode
  // Word 6: # user-addressable sectors per track in default xlate mode
  // Note: words 4,5 are now "Vendor specific (obsolete)"
  BX_HD_THIS s.id_drive[4] = (512 * BX_HD_THIS s.hard_drive.sectors);
  BX_HD_THIS s.id_drive[5] = 512;
  BX_HD_THIS s.id_drive[6] = BX_HD_THIS s.hard_drive.sectors;

  // Word 7-9: Vendor specific
  for (i=7; i<=9; i++)
    BX_HD_THIS s.id_drive[i] = 0;

  // Word 10-19: Serial number (20 ASCII characters, 0000h=not specified)
  // This field is right justified and padded with spaces (20h).
  for (i=10; i<=19; i++)
    BX_HD_THIS s.id_drive[i] = 0;

  // Word 20: buffer type
  //          0000h = not specified
  //          0001h = single ported single sector buffer which is
  //                  not capable of simulataneous data xfers to/from
  //                  the host and the disk.
  //          0002h = dual ported multi-sector buffer capable of
  //                  simulatenous data xfers to/from the host and disk.
  //          0003h = dual ported mutli-sector buffer capable of
  //                  simulatenous data xfers with a read caching
  //                  capability.
  //          0004h-ffffh = reserved
  BX_HD_THIS s.id_drive[20] = 3;

  // Word 21: buffer size in 512 byte increments, 0000h = not specified
  BX_HD_THIS s.id_drive[21] = 512; // 512 Sectors = 256kB cache

  // Word 22: # of ECC bytes available on read/write long cmds
  //          0000h = not specified
  BX_HD_THIS s.id_drive[22] = 4;

  // Word 23..26: Firmware revision (8 ascii chars, 0000h=not specified)
  // This field is left justified and padded with spaces (20h)
  for (i=23; i<=26; i++)
    BX_HD_THIS s.id_drive[i] = 0;

  // Word 27..46: Model number (40 ascii chars, 0000h=not specified)
  // This field is left justified and padded with spaces (20h)
//  for (i=27; i<=46; i++)
//    BX_HD_THIS s.id_drive[i] = 0;
  for (i=0; i<20; i++) {
    BX_HD_THIS s.id_drive[27+i] = (model_no[i*2] << 8) |
                                  model_no[i*2 + 1];
    }

  // Word 47: 15-8 Vendor unique
  //           7-0 00h= read/write multiple commands not implemented
  //               xxh= maximum # of sectors that can be transferred
  //                    per interrupt on read and write multiple commands
  BX_HD_THIS s.id_drive[47] = max_multiple_sectors;

  // Word 48: 0000h = cannot perform dword IO
  //          0001h = can    perform dword IO
  BX_HD_THIS s.id_drive[48] = 0;

  // Word 49: Capabilities
  //   15-10: 0 = reserved
  //       9: 1 = LBA supported
  //       8: 1 = DMA supported
  //     7-0: Vendor unique
  BX_HD_THIS s.id_drive[49] = 0;

  // Word 50: Reserved
  BX_HD_THIS s.id_drive[50] = 0;

  // Word 51: 15-8 PIO data transfer cycle timing mode
  //           7-0 Vendor unique
  BX_HD_THIS s.id_drive[51] = 0x200;

  // Word 52: 15-8 DMA data transfer cycle timing mode
  //           7-0 Vendor unique
  BX_HD_THIS s.id_drive[52] = 0x200;

  // Word 53: 15-1 Reserved
  //             0 1=the fields reported in words 54-58 are valid
  //               0=the fields reported in words 54-58 may be valid
  BX_HD_THIS s.id_drive[53] = 0;

  // Word 54: # of user-addressable cylinders in curr xlate mode
  // Word 55: # of user-addressable heads in curr xlate mode
  // Word 56: # of user-addressable sectors/track in curr xlate mode
  BX_HD_THIS s.id_drive[54] = BX_HD_THIS s.hard_drive.cylinders;
  BX_HD_THIS s.id_drive[55] = BX_HD_THIS s.hard_drive.heads;
  BX_HD_THIS s.id_drive[56] = BX_HD_THIS s.hard_drive.sectors;

  // Word 57-58: Current capacity in sectors
  // Excludes all sectors used for device specific purposes.
  temp32 = 
    BX_HD_THIS s.hard_drive.cylinders *
    BX_HD_THIS s.hard_drive.heads *
    BX_HD_THIS s.hard_drive.sectors;
  BX_HD_THIS s.id_drive[57] = (temp32 & 0xffff); // LSW
  BX_HD_THIS s.id_drive[58] = (temp32 >> 16);    // MSW

  // Word 59: 15-9 Reserved
  //             8 1=multiple sector setting is valid
  //           7-0 current setting for number of sectors that can be
  //               transferred per interrupt on R/W multiple commands
  BX_HD_THIS s.id_drive[59] = 0x0000 | curr_multiple_sectors;

  // Word 60-61:
  // If drive supports LBA Mode, these words reflect total # of user
  // addressable sectors.  This value does not depend on the current
  // drive geometry.  If the drive does not support LBA mode, these
  // words shall be set to 0.
  BX_HD_THIS s.id_drive[60] = 0x0; // LSW
  BX_HD_THIS s.id_drive[61] = 0x0; // MSW

  // Word 62: 15-8 single word DMA transfer mode active
  //           7-0 single word DMA transfer modes supported
  // The low order byte identifies by bit, all the Modes which are
  // supported e.g., if Mode 0 is supported bit 0 is set.
  // The high order byte contains a single bit set to indiciate
  // which mode is active.
  BX_HD_THIS s.id_drive[62] = 0x0;

  // Word 63: 15-8 multiword DMA transfer mode active
  //           7-0 multiword DMA transfer modes supported
  // The low order byte identifies by bit, all the Modes which are
  // supported e.g., if Mode 0 is supported bit 0 is set.
  // The high order byte contains a single bit set to indiciate
  // which mode is active.
  BX_HD_THIS s.id_drive[63] = 0x0;

  // Word 64-127 Reserved
  for (i=64; i<=127; i++)
    BX_HD_THIS s.id_drive[i] = 0;

  // Word 128-159 Vendor unique
  for (i=128; i<=159; i++)
    BX_HD_THIS s.id_drive[i] = 0;

  // Word 160-255 Reserved
  for (i=160; i<=255; i++)
    BX_HD_THIS s.id_drive[i] = 0;

  if (bx_dbg.disk)
    bx_printf ("disk: Drive ID Info. initialized : %04d \n", 512);

  // now convert the id_drive array (native 256 word format) to
  // the controller buffer (512 bytes)
  for (i=0; i<=255; i++) {
    temp16 = BX_HD_THIS s.id_drive[i];
    BX_HD_THIS s.controller.buffer[i*2] = temp16 & 0x00ff;
    BX_HD_THIS s.controller.buffer[i*2+1] = temp16 >> 8;
    }
}

  void
bx_hard_drive_c::command_aborted(unsigned value)
{
  bx_printf("disk: aborting on command 0x%02x\n", value);
  BX_HD_THIS s.controller.current_command = 0;
  BX_HD_THIS s.controller.status.busy = 0;
  BX_HD_THIS s.controller.status.drive_ready = 1;
  BX_HD_THIS s.controller.status.err = 1;
  BX_HD_THIS s.controller.error_register = 0x04; // command ABORTED
  BX_HD_THIS s.controller.status.drq = 0;
  BX_HD_THIS s.controller.status.seek_complete = 0;
  BX_HD_THIS s.controller.status.corrected_data = 0;
  BX_HD_THIS s.controller.buffer_index = 0;
  if (!BX_HD_THIS s.controller.control.disable_irq)
    BX_HD_THIS devices->pic->trigger_irq(14);
}
