// Copyright Notice
// ================
// BOCHS is Copyright 1994-1999 by Kevin P. Lawton.
//
// BOCHS is commercial software.
//
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
//
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02452-3028
//
//     EMail:    bochs@world.std.com







typedef struct {
  int      fd;
  unsigned cylinders;
  unsigned heads;
  unsigned sectors;
  } hard_drive_t;

typedef struct {
  struct {
    Boolean busy;
    Boolean drive_ready;
    Boolean write_fault;
    Boolean seek_complete;
    Boolean drq;
    Boolean corrected_data;
    Boolean index_pulse;
    unsigned index_pulse_count;
    Boolean err;
    } status;
  Bit8u    error_register;
  Bit8u    drive_select;
  Bit8u    head_no;
  Bit8u    sector_count;
  Bit8u    sector_no;
  Bit16u   cylinder_no;
  Bit8u    buffer[512];
  unsigned buffer_index;
  Bit8u    current_command;
  struct {
    Boolean reset;       // 0=normal, 1=reset controller
    Boolean disable_irq; // 0=allow irq, 1=disable irq
    } control;
  } controller_t;



#if BX_USE_HD_SMF
#  define BX_HD_SMF  static
#  define BX_HD_THIS bx_hard_drive.
#else
#  define BX_HD_SMF
#  define BX_HD_THIS this->
#endif


class bx_hard_drive_c {
public:

  bx_hard_drive_c(void);
  ~bx_hard_drive_c(void);
  BX_HD_SMF void   close_harddrive(void);
  BX_HD_SMF void   init(bx_devices_c *d, bx_cmos_c *cmos);

#if !BX_USE_HD_SMF
  Bit32u read(Bit32u address, unsigned io_len);
  void   write(Bit32u address, Bit32u value, unsigned io_len);
#endif

  static Bit32u read_handler(void *this_ptr, Bit32u address, unsigned io_len);
  static void   write_handler(void *this_ptr, Bit32u address, Bit32u value, unsigned io_len);


private:

  BX_HD_SMF void identify_drive(unsigned drive);
  BX_HD_SMF void command_aborted(unsigned command);

  struct {
    hard_drive_t hard_drive;
    controller_t controller;
    // 512 byte buffer for ID drive command
    // These words are stored in native word endian format, as
    // they are fetched and returned via a return(), so
    // there's no need to keep them in x86 endian format.
    Bit16u id_drive[256];
    } s;

  bx_devices_c *devices;
  };

#if BX_USE_HD_SMF
extern bx_hard_drive_c bx_hard_drive;
#endif
